@@ -20,14 +20,7 @@
 #include "optimizer/clauses.h"
 
 #include "nodes/print.h"
-
-
-#if PG_VERSION_NUM >= 120000
-
 #include "optimizer/optimizer.h"
-
-#endif
-
 #include "parser/parse_node.h"
 
 #if PG_VERSION_NUM >= 140000
@@ -42,7 +35,7 @@
 #include "utils/lsyscache.h"
 
 static void collect_volatility(PLpgSQL_checkstate *cstate, Query *query);
-static Query * ExprGetQuery(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr);
+static Query * ExprGetQuery(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, CachedPlanSource *plansource);
 
 static CachedPlan * get_cached_plan(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool *has_result_desc);
 static void plan_checks(PLpgSQL_checkstate *cstate, CachedPlan *cplan, char *query_str);
@@ -81,9 +74,15 @@ parserhook_wrapper_update_used_variables(ParseState *pstate, Node *node)
 			int			dno = p->paramid - 1;
 
 			expr = (PLpgSQL_expr *) pstate->p_ref_hook_state;
+
+			Assert(expr);
+
 			cstate = expr->func->cur_estate->plugin_info;
 
-			if (expr && bms_is_member(dno, expr->paramnos))
+			if (!cstate || cstate->ci_magic != CI_MAGIC)
+				return;
+
+			if (bms_is_member(dno, expr->paramnos))
 			{
 
 #if PG_VERSION_NUM >= 140000
@@ -165,20 +164,14 @@ plpgsql_parser_setup_wrapper(struct ParseState *pstate, PLpgSQL_expr *expr)
 	pstate->p_paramref_hook = param_ref;
 }
 
-/*
- * Generate a prepared plan - this is simplified copy from pl_exec.c Is not
- * necessary to check simple plan, returns true, when expression is
- * succesfully prepared.
- */
 static void
-prepare_plan(PLpgSQL_checkstate *cstate,
+_prepare_plan(PLpgSQL_checkstate *cstate,
 			 PLpgSQL_expr *expr,
 			 int cursorOptions,
 			 ParserSetupHook parser_setup,
 			 void *arg)
 {
 	SPIPlanPtr	plan;
-	Query	   *query;
 
 	if (expr->plan == NULL)
 	{
@@ -228,6 +221,7 @@ prepare_plan(PLpgSQL_checkstate *cstate,
 
 #endif
 
+			expr->func->cur_estate->plugin_info = old_plugin_info;
 		}
 		PG_CATCH();
 		{
@@ -274,8 +268,37 @@ prepare_plan(PLpgSQL_checkstate *cstate,
 
 		SPI_freeplan(plan);
 	}
+}
 
-	query = ExprGetQuery(cstate, expr);
+/*
+ * Generate a prepared plan - this is simplified copy from pl_exec.c Is not
+ * necessary to check simple plan. The planning is forced when plancache is
+ * not valid.
+ */
+static void
+prepare_plan(PLpgSQL_checkstate *cstate,
+			 PLpgSQL_expr *expr,
+			 int cursorOptions,
+			 ParserSetupHook parser_setup,
+			 void *arg)
+{
+	Query	   *query;
+	CachedPlanSource *plansource = NULL;
+
+	do
+	{
+		_prepare_plan(cstate, expr, cursorOptions, parser_setup, arg);
+		plansource = plpgsql_check_get_plan_source(cstate, expr->plan);
+		if (!plansource)
+			return;
+		if (!plansource->is_valid)
+			expr->plan = NULL;
+	}
+	while (!plansource->is_valid);
+
+	query = ExprGetQuery(cstate, expr, plansource);
+	if (!query)
+		return;
 
 	/* there checks are common on every expr/query */
 	plpgsql_check_funcexpr(cstate, query, expr->query);
@@ -331,13 +354,15 @@ CachedPlanSource *
 plpgsql_check_get_plan_source(PLpgSQL_checkstate *cstate, SPIPlanPtr plan)
 {
 	CachedPlanSource *plansource = NULL;
+	int			nplans;
 
 	if (plan == NULL || plan->magic != _SPI_PLAN_MAGIC)
 		elog(ERROR, "cached plan is not valid plan");
 
 	cstate->has_mp = false;
 
-	if (list_length(plan->plancache_list) != 1)
+	nplans = list_length(plan->plancache_list);
+	if (nplans > 1)
 	{
 		/*
 		 * We can allow multiple plans for commands executed by
@@ -353,7 +378,7 @@ plpgsql_check_get_plan_source(PLpgSQL_checkstate *cstate, SPIPlanPtr plan)
 		else
 			elog(ERROR, "plan is not single execution plany");
 	}
-	else
+	else if (nplans == 1)
 		plansource = (CachedPlanSource *) linitial(plan->plancache_list);
 
 	return plansource;
@@ -364,12 +389,15 @@ plpgsql_check_get_plan_source(PLpgSQL_checkstate *cstate, SPIPlanPtr plan)
  *
  */
 static Query *
-ExprGetQuery(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
+ExprGetQuery(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, CachedPlanSource *plansource)
 {
-	CachedPlanSource *plansource;
 	Query *result = NULL;
 
-	plansource = plpgsql_check_get_plan_source(cstate, expr->plan);
+	Assert(plansource);
+	Assert(plansource->is_valid);
+
+	if (!plansource->query_list)
+		elog(ERROR, "missing plan in plancache source");
 
 	/*
 	 * query_list has more fields, when rules are used. There
@@ -524,6 +552,11 @@ get_cached_plan(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool *has_result
 	CachedPlan	*cplan;
 
 	plansource = plpgsql_check_get_plan_source(cstate, expr->plan);
+	if (!plansource)
+	{
+		*has_result_desc = false;
+		return NULL;
+	}
 
 	*has_result_desc = plansource->resultDesc ? true : false;
 
@@ -699,11 +732,13 @@ plpgsql_check_expr_get_node(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool
 	if (has_result_desc && IsA(_stmt, PlannedStmt) &&_stmt->commandType == CMD_SELECT)
 	{
 		Plan	   *_plan;
-		TargetEntry *tle;
 
 		_plan = _stmt->planTree;
-		if (IsA(_plan, Result) &&list_length(_plan->targetlist) == 1)
+
+		if ((IsA(_plan, Result) || IsA(_plan, ProjectSet)) && list_length(_plan->targetlist) == 1)
 		{
+			TargetEntry *tle;
+
 			tle = (TargetEntry *) linitial(_plan->targetlist);
 			result = (Node *) tle->expr;
 		}
@@ -754,10 +789,15 @@ is_const_null_expr(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 }
 
 char *
-plpgsql_check_const_to_string(Const *c)
+plpgsql_check_const_to_string(Node *node, int *location)
 {
-	if (IsA((Node *) c, Const))
+	if (IsA(node, Const))
 	{
+		Const *c = (Const *) node;
+
+		if (location)
+			*location = c->location;
+
 		if (!c->constisnull)
 		{
 			Oid		typoutput;
@@ -772,25 +812,72 @@ plpgsql_check_const_to_string(Const *c)
 	return NULL;
 }
 
+char *
+plpgsql_check_get_tracked_const(PLpgSQL_checkstate *cstate, Node *node)
+{
+	if (!cstate->strconstvars)
+		return NULL;
+
+	if (cstate->pragma_vector.disable_constants_tracing)
+		return NULL;
+
+	if (IsA(node, Param))
+	{
+		Param *p = (Param *) node;
+
+		if (p->paramkind == PARAM_EXTERN && p->paramid > 0 && p->location != -1)
+		{
+			int			dno = p->paramid - 1;
+
+			if (cstate->strconstvars[dno])
+				return cstate->strconstvars[dno];
+		}
+	}
+	else if (IsA(node, CoerceViaIO))
+	{
+		CoerceViaIO *coerce = (CoerceViaIO *) node;
+		bool	typispreferred;
+		char 	typcategory;
+
+		get_type_category_preferred(coerce->resulttype, &typcategory, &typispreferred);
+		if (typcategory == 'S')
+			return plpgsql_check_get_tracked_const(cstate, (Node *) coerce->arg);
+	}
+
+	return NULL;
+}
+
+char *
+plpgsql_check_get_const_string(PLpgSQL_checkstate *cstate, Node *node, int *location)
+{
+	char	   *str;
+
+	if (location)
+		*location = -1;
+
+	str = plpgsql_check_const_to_string(node, location);
+
+	if (!str)
+		str = plpgsql_check_get_tracked_const(cstate, node);
+
+	return str;
+}
+
 /*
  * Returns string for any not null constant. isnull is true,
  * when constant is null.
  *
  */
 char *
-plpgsql_check_expr_get_string(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool *isnull)
+plpgsql_check_expr_get_string(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, int *location)
 {
-	Const	   *c;
 
-	c = expr_get_const(cstate, expr);
-	if (c)
-	{
-		*isnull = c->constisnull;
+	Node *node = plpgsql_check_expr_get_node(cstate, expr, true);
 
-		return plpgsql_check_const_to_string(c);
-	}
+	if (!node)
+		return NULL;
 
-	return NULL;
+	return plpgsql_check_get_const_string(cstate, node, location);
 }
 
 static void
@@ -800,6 +887,8 @@ force_plan_checks(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 	bool		has_result_desc;
 
 	cplan = get_cached_plan(cstate, expr, &has_result_desc);
+	if (!cplan)
+		return;
 
 	/* do all checks for this plan, reduce a access to plan cache */
 	plan_checks(cstate, cplan, expr->query);
@@ -889,15 +978,13 @@ plpgsql_check_expr_with_scalar_type(PLpgSQL_checkstate *cstate,
 								    expected_typoid, -1,
 								    TupleDescAttr(tupdesc, 0)->atttypid,
 								    is_immutable_null);
-		}
 
-		ReleaseTupleDesc(tupdesc);
+			ReleaseTupleDesc(tupdesc);
+		}
 
 		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
-
-		SPI_restore_connection();
 	}
 	PG_CATCH();
 	{
@@ -921,9 +1008,6 @@ plpgsql_check_expr_with_scalar_type(PLpgSQL_checkstate *cstate,
 		else
 			plpgsql_check_put_error_edata(cstate, edata);
 		MemoryContextSwitchTo(oldCxt);
-
-		/* reconnect spi */
-		SPI_restore_connection();
 	}
 	PG_END_TRY();
 }
@@ -1044,8 +1128,6 @@ plpgsql_check_returned_expr(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool
 		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
-
-		SPI_restore_connection();
 	}
 	PG_CATCH();
 	{
@@ -1069,13 +1151,42 @@ plpgsql_check_returned_expr(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool
 		else
 			plpgsql_check_put_error_edata(cstate, edata);
 		MemoryContextSwitchTo(oldCxt);
-
-		/* reconnect spi */
-		SPI_restore_connection();
 	}
 	PG_END_TRY();
 }
 
+/*
+ * Release all string constants related to variables from row variable
+ */
+static void
+free_string_constant(PLpgSQL_checkstate *cstate, PLpgSQL_row *row)
+{
+	int			fnum;
+
+	for (fnum = 0; fnum < row->nfields; fnum++)
+	{
+		PLpgSQL_datum *field;
+		int				varno = row->varnos[fnum];
+
+		/* skip dropped fields */
+		if (varno < 0)
+			continue;
+
+		/* the assigned value is not constant, reset current */
+		if (cstate->strconstvars && cstate->strconstvars[varno])
+		{
+			pfree(cstate->strconstvars[varno]);
+			cstate->strconstvars[varno] = NULL;
+		}
+
+		field = cstate->estate->datums[fnum];
+
+		if (field->dtype == PLPGSQL_DTYPE_ROW)
+			free_string_constant(cstate, (PLpgSQL_row *) field);
+	}
+}
+
+
 /*
  * Check expression as rvalue - on right in assign statement. It is used for
  * only expression check too - when target is unknown.
@@ -1268,12 +1379,12 @@ plpgsql_check_expr_as_rvalue(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr,
 		if (cstate->cinfo->compatibility_warnings && targetdno != -1)
 		{
 			PLpgSQL_var *var = (PLpgSQL_var *) cstate->estate->datums[targetdno];
-			bool		is_ok = true;
 
 			if (var->dtype == PLPGSQL_DTYPE_VAR && var->datatype->typoid == REFCURSOROID)
 			{
 				Node *node = plpgsql_check_expr_get_node(cstate, expr, false);
 				bool		is_declare_cursor;
+				bool		is_ok = true;
 
 				is_declare_cursor = cstate->estate->err_stmt &&
 									cstate->estate->err_stmt->cmd_type == PLPGSQL_STMT_BLOCK &&
@@ -1339,6 +1450,56 @@ plpgsql_check_expr_as_rvalue(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr,
 			}
 		}
 
+		if (cstate->cinfo->constants_tracing && targetrow)
+		{
+			/*
+			 * We cannot to cut constants with results now, but we have to
+			 * reset all possible string constants saved in variables from
+			 * row variable.
+			 */
+			free_string_constant(cstate, targetrow);
+		}
+		else if (cstate->cinfo->constants_tracing && targetdno != -1)
+		{
+			char	   *str;
+
+			str = plpgsql_check_expr_get_string(cstate, expr, NULL);
+			if (str)
+			{
+				PLpgSQL_stmt_stack_item *current = cstate->top_stmt_stack;
+				MemoryContext oldcxt = MemoryContextSwitchTo(cstate->check_cxt);
+				char	   *prev_val;
+
+				Assert(cstate->top_stmt_stack);
+
+				if (!cstate->strconstvars)
+					cstate->strconstvars = palloc0(sizeof(char *) * cstate->estate->ndatums);
+
+				/*
+				 * We need to do copy string first. There is an possibility to self
+				 * reference, and then we need to first copy, and after that free.
+				 */
+				prev_val = cstate->strconstvars[targetdno];
+				cstate->strconstvars[targetdno] = pstrdup(str);
+
+				if (prev_val)
+					pfree(prev_val);
+
+				current->invalidate_strconstvars = bms_add_member(current->invalidate_strconstvars, targetdno);
+
+				MemoryContextSwitchTo(oldcxt);
+			}
+			else
+			{
+				/* the assigned value is not constant, reset current */
+				if (cstate->strconstvars && cstate->strconstvars[targetdno])
+				{
+					pfree(cstate->strconstvars[targetdno]);
+					cstate->strconstvars[targetdno] = NULL;
+				}
+			}
+		}
+
 		if (expected_typoid != InvalidOid && type_is_rowtype(expected_typoid) && first_level_typoid != InvalidOid)
 		{
 			/* simple error, scalar source to composite target */
@@ -1397,8 +1558,6 @@ no_other_check:
 		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
-
-		SPI_restore_connection();
 	}
 	PG_CATCH();
 	{
@@ -1422,9 +1581,6 @@ no_other_check:
 		else
 			plpgsql_check_put_error_edata(cstate, edata);
 		MemoryContextSwitchTo(oldCxt);
-
-		/* reconnect spi */
-		SPI_restore_connection();
 	}
 	PG_END_TRY();
 }
@@ -1464,7 +1620,6 @@ plpgsql_check_expr_as_sqlstmt(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 {
 	ResourceOwner oldowner;
 	MemoryContext oldCxt = CurrentMemoryContext;
-	TupleDesc	tupdesc;
 	volatile bool result = false;
 
 	if (!expr)
@@ -1476,6 +1631,8 @@ plpgsql_check_expr_as_sqlstmt(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 
 	PG_TRY();
 	{
+		TupleDesc	tupdesc;
+
 		prepare_plan(cstate, expr, 0, NULL, NULL);
 		/* record all variables used by the query */
 		cstate->used_variables = bms_add_members(cstate->used_variables, expr->paramnos);
@@ -1491,8 +1648,6 @@ plpgsql_check_expr_as_sqlstmt(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
-
-		SPI_restore_connection();
 	}
 	PG_CATCH();
 	{
@@ -1516,9 +1671,6 @@ plpgsql_check_expr_as_sqlstmt(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 		else
 			plpgsql_check_put_error_edata(cstate, edata);
 		MemoryContextSwitchTo(oldCxt);
-
-		/* reconnect spi */
-		SPI_restore_connection();
 	}
 	PG_END_TRY();
 
@@ -1590,8 +1742,6 @@ plpgsql_check_assignment(PLpgSQL_checkstate *cstate,
 								 is_expression);
 }
 
-#if PG_VERSION_NUM >= 110000
-
 void
 plpgsql_check_assignment_to_variable(PLpgSQL_checkstate *cstate,
 							 PLpgSQL_expr *expr,
@@ -1631,9 +1781,6 @@ plpgsql_check_assignment_to_variable(PLpgSQL_checkstate *cstate,
 									 true);
 }
 
-#endif
-
-
 /*
  * row->nfields can cound dropped columns. When this behave can raise
  * false alarms, we should to count fields more precisely.
