@@ -21,13 +21,8 @@
 #include "utils/rel.h"
 #include "utils/syscache.h"
 #include "utils/typcache.h"
-
-#if PG_VERSION_NUM >= 120000
-
 #include "access/heapam.h"
 
-#endif
-
 static HTAB *plpgsql_check_HashTable = NULL;
 
 bool plpgsql_check_other_warnings = false;
@@ -35,6 +30,7 @@ bool plpgsql_check_extra_warnings = false;
 bool plpgsql_check_performance_warnings = false;
 bool plpgsql_check_compatibility_warnings = false;
 bool plpgsql_check_fatal_errors = true;
+bool plpgsql_check_constants_tracing = true;
 int plpgsql_check_mode = PLPGSQL_CHECK_MODE_BY_FUNCTION;
 
 /* ----------
@@ -61,6 +57,13 @@ static void setup_estate(PLpgSQL_execstate *estate, PLpgSQL_function *func, Retu
 static void setup_cstate(PLpgSQL_checkstate *cstate, plpgsql_check_result_info *result_info,
 	plpgsql_check_info *cinfo, bool is_active_mode, bool fake_rtd);
 
+static void passive_check_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func, void **plugin2_info);
+
+static plpgsql_check_plugin2 check_plugin2 = { NULL,
+											   passive_check_func_beg, NULL, NULL,
+											   NULL, NULL, NULL,
+											   NULL, NULL, NULL, NULL, NULL };
+
 /*
  * Prepare list of OUT variables for later report
  */
@@ -114,20 +117,9 @@ plpgsql_check_function_internal(plpgsql_check_result_info *ri,
 {
 	PLpgSQL_checkstate cstate;
 	PLpgSQL_function *volatile function = NULL;
-	int			save_nestlevel = 0;
 	bool		reload_config;
-
-#if PG_VERSION_NUM >= 120000
-
 	LOCAL_FCINFO(fake_fcinfo, 0);
 
-#else
-
-	FunctionCallInfoData fake_fcinfo_data;
-	FunctionCallInfo fake_fcinfo = &fake_fcinfo_data;
-
-#endif
-
 	FmgrInfo	flinfo;
 	TriggerData trigdata;
 	EventTriggerData etrigdata;
@@ -169,13 +161,14 @@ plpgsql_check_function_internal(plpgsql_check_result_info *ri,
 		Oid		   *argtypes;
 		char	  **argnames;
 		char	   *argmodes;
-		int			i;
 
 		numargs = get_func_arg_info(cinfo->proctuple,
 							&argtypes, &argnames, &argmodes);
 
 		if (argnames != NULL)
 		{
+			int			i;
+
 			for (i = 0; i < numargs; i++)
 			{
 				if (argnames[i][0] != '\0')
@@ -188,6 +181,8 @@ plpgsql_check_function_internal(plpgsql_check_result_info *ri,
 
 	PG_TRY();
 	{
+		int			save_nestlevel = 0;
+
 		BeginInternalSubTransaction(NULL);
 		MemoryContextSwitchTo(cstate.check_cxt);
 
@@ -254,8 +249,6 @@ plpgsql_check_function_internal(plpgsql_check_result_info *ri,
 			relation_close(trigdata.tg_relation, AccessShareLock);
 
 		release_exprs(cstate.exprs);
-
-		SPI_restore_connection();
 	}
 	PG_CATCH();
 	{
@@ -280,9 +273,6 @@ plpgsql_check_function_internal(plpgsql_check_result_info *ri,
 		}
 
 		plpgsql_check_put_error_edata(&cstate, edata);
-
-		/* reconnect spi */
-		SPI_restore_connection();
 	}
 	PG_END_TRY();
 
@@ -303,18 +293,13 @@ plpgsql_check_function_internal(plpgsql_check_result_info *ri,
  *      and local variables are initialized.
  *
  */
-void
-plpgsql_check_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func)
+static void
+passive_check_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func, void **plugin2_info)
 {
 	const char *err_text = estate->err_text;
 	int closing;
 	List		*exceptions;
 
-	plpgsql_check_profiler_func_beg(estate, func);
-
-	if (plpgsql_check_tracer)
-		plpgsql_check_tracer_on_func_beg(estate, func);
-
 	if (plpgsql_check_mode == PLPGSQL_CHECK_MODE_FRESH_START ||
 		   plpgsql_check_mode == PLPGSQL_CHECK_MODE_EVERY_START)
 	{
@@ -363,6 +348,7 @@ plpgsql_check_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func)
 		cinfo.performance_warnings = plpgsql_check_performance_warnings,
 		cinfo.extra_warnings = plpgsql_check_extra_warnings,
 		cinfo.compatibility_warnings = plpgsql_check_compatibility_warnings;
+		cinfo.constants_tracing = plpgsql_check_constants_tracing;
 
 		ri.format = PLPGSQL_CHECK_FORMAT_ELOG;
 
@@ -388,8 +374,6 @@ plpgsql_check_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func)
 			{
 				PLpgSQL_rec *rec = (PLpgSQL_rec *) estate->datums[i];
 
-#if PG_VERSION_NUM >= 110000
-
 				memcpy(&saved_records[i], rec, sizeof(PLpgSQL_rec));
 
 				if (rec->erh)
@@ -398,19 +382,6 @@ plpgsql_check_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func)
 					rec->erh = make_expanded_record_from_exprecord(rec->erh, cstate.check_cxt);
 				}
 
-#else
-
-				saved_records[i].tup = rec->tup;
-				saved_records[i].tupdesc = rec->tupdesc;
-				saved_records[i].freetup = rec->freetup;
-				saved_records[i].freetupdesc = rec->freetupdesc;
-
-				/* don't release a original tupdesc and original tup */
-				rec->freetup = false;
-				rec->freetupdesc = false;
-
-#endif
-
 			}
 			else if (estate->datums[i]->dtype == PLPGSQL_DTYPE_VAR)
 			{
@@ -486,22 +457,7 @@ plpgsql_check_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func)
 			{
 				PLpgSQL_rec *rec = (PLpgSQL_rec *) estate->datums[i];
 
-#if PG_VERSION_NUM >= 110000
-
 				memcpy(rec, &saved_records[i], sizeof(PLpgSQL_rec));
-
-#else
-
-				if (rec->freetupdesc)
-					FreeTupleDesc(rec->tupdesc);
-
-				rec->tup = saved_records[i].tup;
-				rec->tupdesc = saved_records[i].tupdesc;
-				rec->freetup = saved_records[i].freetup;
-				rec->freetupdesc = saved_records[i].freetupdesc;
-
-#endif
-
 			}
 			else if (estate->datums[i]->dtype == PLPGSQL_DTYPE_VAR)
 			{
@@ -518,6 +474,11 @@ plpgsql_check_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func)
 	}
 }
 
+void
+plpgsql_check_passive_check_init(void)
+{
+	plpgsql_check_register_pldbgapi2_plugin(&check_plugin2);
+}
 
 /*
  * Check function - it prepare variables and starts a prepare plan walker
@@ -606,8 +567,6 @@ function_check(PLpgSQL_function *func, PLpgSQL_checkstate *cstate)
 static void
 trigger_check(PLpgSQL_function *func, Node *tdata, PLpgSQL_checkstate *cstate)
 {
-	PLpgSQL_rec *rec_new,
-			   *rec_old;
 	int			i;
 	int closing = PLPGSQL_CHECK_UNCLOSED;
 	List	   *exceptions;
@@ -622,6 +581,9 @@ trigger_check(PLpgSQL_function *func, Node *tdata, PLpgSQL_checkstate *cstate)
 
 	if (IsA(tdata, TriggerData))
 	{
+		PLpgSQL_rec *rec_new,
+				   *rec_old;
+
 		TriggerData *trigdata = (TriggerData *) tdata;
 
 		/*
@@ -634,7 +596,6 @@ trigger_check(PLpgSQL_function *func, Node *tdata, PLpgSQL_checkstate *cstate)
 		 * NEW.foo = 'xyz')", which should parse regardless of the current
 		 * trigger type.
 		 */
-#if PG_VERSION_NUM >= 110000
 
 		/*
 		 * find all PROMISE VARIABLES and initit their
@@ -651,46 +612,11 @@ trigger_check(PLpgSQL_function *func, Node *tdata, PLpgSQL_checkstate *cstate)
 		plpgsql_check_recval_assign_tupdesc(cstate, rec_new, trigdata->tg_relation->rd_att, false);
 		rec_old = (PLpgSQL_rec *) (cstate->estate->datums[func->old_varno]);
 		plpgsql_check_recval_assign_tupdesc(cstate, rec_old, trigdata->tg_relation->rd_att, false);
-
-#else
-
-		rec_new = (PLpgSQL_rec *) (cstate->estate->datums[func->new_varno]);
-		rec_new->freetup = false;
-		rec_new->freetupdesc = false;
-		plpgsql_check_assign_tupdesc_row_or_rec(cstate, NULL, rec_new, trigdata->tg_relation->rd_att, false);
-
-		rec_old = (PLpgSQL_rec *) (cstate->estate->datums[func->old_varno]);
-		rec_old->freetup = false;
-		rec_old->freetupdesc = false;
-		plpgsql_check_assign_tupdesc_row_or_rec(cstate, NULL, rec_old, trigdata->tg_relation->rd_att, false);
-
-		/*
-		 * Assign the special tg_ variables
-		 */
-		init_datum_dno(cstate, func->tg_op_varno, true, true);
-		init_datum_dno(cstate, func->tg_name_varno, true, true);
-		init_datum_dno(cstate, func->tg_when_varno, true, true);
-		init_datum_dno(cstate, func->tg_level_varno, true, true);
-		init_datum_dno(cstate, func->tg_relid_varno, true, true);
-		init_datum_dno(cstate, func->tg_relname_varno, true, true);
-		init_datum_dno(cstate, func->tg_table_name_varno, true, true);
-		init_datum_dno(cstate, func->tg_table_schema_varno, true, true);
-		init_datum_dno(cstate, func->tg_nargs_varno, true, true);
-		init_datum_dno(cstate, func->tg_argv_varno, true, true);
-
-#endif
-
 	}
 	else if (IsA(tdata, EventTriggerData))
 	{
 
-#if PG_VERSION_NUM < 110000
-
-		init_datum_dno(cstate, func->tg_event_varno, true, true);
-		init_datum_dno(cstate, func->tg_tag_varno, true, true);
-
-#endif
-
+		/* do nothing */
 	}
 	else
 		elog(ERROR, "unexpected environment");
@@ -841,16 +767,8 @@ plpgsql_check_setup_fcinfo(plpgsql_check_info *cinfo,
 	*fake_rtd = false;
 
 	/* clean structures */
-#if PG_VERSION_NUM >= 120000
-
 	MemSet(fcinfo, 0, SizeForFunctionCallInfo(0));
 
-#else
-
-	MemSet(fcinfo, 0, sizeof(FunctionCallInfoData));
-
-#endif
-
 	MemSet(flinfo, 0, sizeof(FmgrInfo));
 	MemSet(rsinfo, 0, sizeof(ReturnSetInfo));
 
@@ -1043,16 +961,8 @@ plpgsql_check_setup_fcinfo(plpgsql_check_info *cinfo,
 		{
 			*fake_rtd = cinfo->rettype == RECORDOID;
 
-#if PG_VERSION_NUM >= 120000
-
 			resultTupleDesc = CreateTemplateTupleDesc(1);
 
-#else
-
-			resultTupleDesc = CreateTemplateTupleDesc(1, false);
-
-#endif
-
 			TupleDescInitEntry(resultTupleDesc,
 							    (AttrNumber) 1, "__result__",
 							    cinfo->rettype, -1, 0);
@@ -1117,21 +1027,12 @@ setup_estate(PLpgSQL_execstate *estate,
 
 	estate->readonly_func = func->fn_readonly;
 
-#if PG_VERSION_NUM < 110000
-
-	estate->rettupdesc = NULL;
-	estate->eval_econtext = NULL;
-
-#else
-
 	estate->eval_econtext = makeNode(ExprContext);
 	estate->eval_econtext->ecxt_per_tuple_memory = AllocSetContextCreate(CurrentMemoryContext,
 													"ExprContext",
 													ALLOCSET_DEFAULT_SIZES);
 	estate->datum_context = CurrentMemoryContext;
 
-#endif
-
 	estate->exitlabel = NULL;
 	estate->cur_error = NULL;
 
@@ -1141,17 +1042,7 @@ setup_estate(PLpgSQL_execstate *estate,
 		estate->tuple_store_cxt = rsi->econtext->ecxt_per_query_memory;
 		estate->tuple_store_owner = CurrentResourceOwner;
 
-#if PG_VERSION_NUM >= 110000
-
 		estate->tuple_store_desc = rsi->expectedDesc;
-
-#else
-
-		if (estate->retisset)
-			estate->rettupdesc = rsi->expectedDesc;
-
-#endif
-
 	}
 	else
 	{
@@ -1168,12 +1059,6 @@ setup_estate(PLpgSQL_execstate *estate,
 	estate->eval_tuptable = NULL;
 	estate->eval_processed = 0;
 
-#if PG_VERSION_NUM < 120000
-
-	estate->eval_lastoid = InvalidOid;
-
-#endif
-
 	if (cinfo->oldtable)
 	{
 		EphemeralNamedRelation enr = palloc(sizeof(EphemeralNamedRelationData));
@@ -1223,6 +1108,8 @@ setup_cstate(PLpgSQL_checkstate *cstate,
 			 bool is_active_mode,
 			 bool fake_rtd)
 {
+	cstate->ci_magic = CI_MAGIC;
+
 	cstate->decl_volatility = cinfo->volatility;
 	cstate->has_execute_stmt = false;
 	cstate->volatility = PROVOLATILE_IMMUTABLE;
@@ -1251,22 +1138,10 @@ setup_cstate(PLpgSQL_checkstate *cstate,
 	cstate->func_oids = NULL;
 	cstate->rel_oids = NULL;
 
-#if PG_VERSION_NUM >= 110000
-
 	cstate->check_cxt = AllocSetContextCreate(CurrentMemoryContext,
 										 "plpgsql_check temporary cxt",
 										   ALLOCSET_DEFAULT_SIZES);
 
-#else
-
-	cstate->check_cxt = AllocSetContextCreate(CurrentMemoryContext,
-										 "plpgsql_check temporary cxt",
-										   ALLOCSET_DEFAULT_MINSIZE,
-										   ALLOCSET_DEFAULT_INITSIZE,
-										   ALLOCSET_DEFAULT_MAXSIZE);
-
-#endif
-
 	cstate->found_return_query = false;
 	cstate->found_return_dyn_query = false;
 
@@ -1281,15 +1156,18 @@ setup_cstate(PLpgSQL_checkstate *cstate,
 	cstate->allow_mp = false;
 
 	cstate->pragma_vector.disable_check = false;
-	cstate->pragma_vector.disable_tracer = false;
 	cstate->pragma_vector.disable_other_warnings = false;
 	cstate->pragma_vector.disable_performance_warnings = false;
 	cstate->pragma_vector.disable_extra_warnings = false;
 	cstate->pragma_vector.disable_security_warnings = false;
 	cstate->pragma_vector.disable_compatibility_warnings = false;
+	cstate->pragma_vector.disable_constants_tracing = false;
 
 	/* try to find oid of plpgsql_check pragma function */
 	cstate->pragma_foid = plpgsql_check_pragma_func_oid();
+
+	/* for simple string constants tracing */
+	cstate->strconstvars = NULL;
 }
 
 /*
@@ -1322,13 +1200,6 @@ load_configuration(HeapTuple procTuple, bool *reload_config)
 			*reload_config = true;
 
 			ProcessGUCArray(set_items,
-
-#if PG_VERSION_NUM >= 160000
-
-							NULL,
-
-#endif
-
 							(superuser() ? PGC_SUSET : PGC_USERSET),
 							PGC_S_SESSION,
 							GUC_ACTION_SAVE);
@@ -1365,13 +1236,7 @@ init_datum_dno(PLpgSQL_checkstate *cstate, int dno, bool is_auto, bool is_protec
 {
 	switch (cstate->estate->datums[dno]->dtype)
 	{
-
-#if PG_VERSION_NUM >= 110000
-
 		case PLPGSQL_DTYPE_PROMISE:
-
-#endif
-
 		case PLPGSQL_DTYPE_VAR:
 			{
 				PLpgSQL_var *var = (PLpgSQL_var *) cstate->estate->datums[dno];
@@ -1382,8 +1247,6 @@ init_datum_dno(PLpgSQL_checkstate *cstate, int dno, bool is_auto, bool is_protec
 			}
 			break;
 
-#if PG_VERSION_NUM >= 110000
-
 		case PLPGSQL_DTYPE_REC:
 			{
 				PLpgSQL_rec *rec = (PLpgSQL_rec *) cstate->estate->datums[dno];
@@ -1393,8 +1256,6 @@ init_datum_dno(PLpgSQL_checkstate *cstate, int dno, bool is_auto, bool is_protec
 			}
 			break;
 
-#endif
-
 		case PLPGSQL_DTYPE_ROW:
 			{
 				PLpgSQL_row *row = (PLpgSQL_row *) cstate->estate->datums[dno];
@@ -1432,13 +1293,7 @@ copy_plpgsql_datum(PLpgSQL_checkstate *cstate, PLpgSQL_datum *datum)
 	switch (datum->dtype)
 	{
 		case PLPGSQL_DTYPE_VAR:
-
-#if PG_VERSION_NUM >= 110000
-
 		case PLPGSQL_DTYPE_PROMISE:
-
-#endif
-
 			{
 				PLpgSQL_var *new = palloc(sizeof(PLpgSQL_var));
 
@@ -1537,12 +1392,12 @@ plpgsql_check_is_checked(PLpgSQL_function *func)
 void
 plpgsql_check_mark_as_checked(PLpgSQL_function *func)
 {
-	plpgsql_check_HashEnt *hentry;
-	bool		found;
-
 	/* don't try to mark anonymous code blocks */
 	if (func->fn_oid != InvalidOid)
 	{
+		plpgsql_check_HashEnt *hentry;
+		bool		found;
+
 		hentry = (plpgsql_check_HashEnt *) hash_search(plpgsql_check_HashTable,
 												 (void *) func->fn_hashkey,
 												 HASH_ENTER,
