@@ -73,8 +73,6 @@ detect_dependency_walker(Node *node, void *context)
 			}
 		}
 
-#if PG_VERSION_NUM >= 110000
-
 		if (query->utilityStmt && IsA(query->utilityStmt, CallStmt))
 		{
 			CallStmt *callstmt = (CallStmt *) query->utilityStmt;
@@ -82,8 +80,6 @@ detect_dependency_walker(Node *node, void *context)
 			detect_dependency_walker((Node *) callstmt->funcexpr, context);
 		}
 
-#endif
-
 		return query_tree_walker((Query *) node,
 								 detect_dependency_walker,
 								 context, 0);
@@ -97,31 +93,26 @@ detect_dependency_walker(Node *node, void *context)
 		{
 			if (!bms_is_member(fexpr->funcid, cstate->func_oids))
 			{
-				StringInfoData		str;
-				ListCell		   *lc;
-				int		i = 0;
-
-#if PG_VERSION_NUM >= 110000
-
+				StringInfoData	str;
+				ListCell   *lc;
+				bool		is_first = true;
 				char		prokind = get_func_prokind(fexpr->funcid);
 
-#endif
-
 				initStringInfo(&str);
 				appendStringInfoChar(&str, '(');
 				foreach(lc, fexpr->args)
 				{
 					Node *expr = (Node *) lfirst(lc);
 
-					if (i++ > 0)
+					if (!is_first)
 						appendStringInfoChar(&str, ',');
+					else
+						is_first = false;
 
 					appendStringInfoString(&str, format_type_be(exprType(expr)));
 				}
 				appendStringInfoChar(&str, ')');
 
-#if PG_VERSION_NUM >= 110000
-
 				plpgsql_check_put_dependency(ri,
 											  prokind == PROKIND_PROCEDURE ? "PROCEDURE" : "FUNCTION",
 											  fexpr->funcid,
@@ -129,17 +120,6 @@ detect_dependency_walker(Node *node, void *context)
 											  get_func_name(fexpr->funcid),
 											  str.data);
 
-#else
-
-				plpgsql_check_put_dependency(ri,
-											  "FUNCTION",
-											  fexpr->funcid,
-											  get_namespace_name(get_func_namespace(fexpr->funcid)),
-											  get_func_name(fexpr->funcid),
-											  str.data);
-
-#endif
-
 				pfree(str.data);
 
 				cstate->func_oids = bms_add_member(cstate->func_oids, fexpr->funcid);
@@ -214,7 +194,6 @@ check_funcexpr_walker(Node *node, void *context)
 	if (IsA(node, FuncExpr))
 	{
 		FuncExpr *fexpr = (FuncExpr *) node;
-		int		location = -1;
 
 		switch (fexpr->funcid)
 		{
@@ -224,8 +203,7 @@ check_funcexpr_walker(Node *node, void *context)
 			case SETVAL2_OID:
 				{
 					Node *first_arg = linitial(fexpr->args);
-
-					location = fexpr->location;
+					int			location = fexpr->location;
 
 					if (first_arg && IsA(first_arg, Const))
 					{
@@ -523,7 +501,7 @@ text_format_parse_format(const char *start_ptr,
 }
 
 #define TOO_FEW_ARGUMENTS_CHECK(arg, nargs) \
-	if (arg > nargs) \
+	if ((arg) > (nargs)) \
 	{ \
 		if (wp) \
 			plpgsql_check_put_error(wp->cstate, \
@@ -538,6 +516,160 @@ text_format_parse_format(const char *start_ptr,
 		return -1; \
 	}
 
+/*
+ * Returns string based on expected result of "format" function.
+ * Try to use traced constant, when are available.
+ */
+char *
+plpgsql_check_get_formatted_string(PLpgSQL_checkstate *cstate,
+								   const char *fmt,
+								   List *args,
+								   bool *found_ident_placeholder,
+								   bool *found_literal_placeholder,
+								   bool *expr_is_const)
+{
+	StringInfoData sinfo;
+	const char	   *cp;
+	const char	   *end_ptr = fmt + strlen(fmt);
+	int			nargs = list_length(args);
+	int			arg = 1;
+	int			_arg;
+	bool		is_error;
+
+	*found_ident_placeholder = false;
+	*found_literal_placeholder = false;
+	*expr_is_const = true;
+
+	initStringInfo(&sinfo);
+
+	/* Scan format string, looking for conversion specifiers. */
+	for (cp = fmt; cp < end_ptr; cp++)
+	{
+		int			argpos;
+		int			widthpos;
+
+		if (*cp != '%')
+		{
+			appendStringInfoChar(&sinfo, *cp);
+			continue;
+		}
+
+		if (++cp >= end_ptr)
+		{
+			pfree(sinfo.data);
+			return NULL;
+		}
+
+		if (*cp == '%')
+		{
+			appendStringInfoChar(&sinfo, '%');
+			continue;
+		}
+
+		/* Parse the optional portions of the format specifier */
+		cp = text_format_parse_format(cp, end_ptr,
+									  &argpos, &widthpos,
+									  -1, NULL, &is_error);
+
+		if (is_error || strchr("sIL", *cp) == NULL)
+		{
+			pfree(sinfo.data);
+			return NULL;
+		}
+
+		if (widthpos >= 0)
+		{
+			if (widthpos > 0)
+			{
+				if (widthpos > nargs)
+				{
+					pfree(sinfo.data);
+					return NULL;
+				}
+			}
+			else
+			{
+				if (++arg > nargs)
+				{
+					pfree(sinfo.data);
+					return NULL;
+				}
+			}
+		}
+
+		_arg = argpos >= 1 ? argpos : arg + 1;
+		if (_arg <= nargs)
+		{
+			char	   *str;
+
+			str = plpgsql_check_get_const_string(cstate, list_nth(args, _arg - 1), NULL);
+
+			if (*cp == 'I')
+			{
+				if (!str)
+				{
+					appendStringInfoString(&sinfo, "\"%I\"");
+					*found_ident_placeholder = true;
+					*expr_is_const = false;
+				}
+				else
+					appendStringInfoString(&sinfo, quote_identifier(str));
+			}
+			else if (*cp == 'L')
+			{
+				if (!str)
+				{
+					/*
+					 * Original idea was used external parameter,
+					 * but external parameters requires known type,
+					 * so most safe value is NULL instead.
+					 */
+					appendStringInfoString(&sinfo, " null ");
+					*found_literal_placeholder = true;
+					*expr_is_const = false;
+				}
+				else
+				{
+					char	   *qstr = quote_literal_cstr(str);
+
+					appendStringInfoString(&sinfo, qstr);
+					pfree(qstr);
+				}
+			}
+			else
+			{
+				if (!str)
+				{
+					pfree(sinfo.data);
+					*expr_is_const = false;
+					return NULL;
+				}
+				else
+					appendStringInfoString(&sinfo, str);
+			}
+		}
+
+		if (argpos >= 1)
+		{
+			if (argpos > nargs)
+			{
+				pfree(sinfo.data);
+				return NULL;
+			}
+		}
+		else
+		{
+			if (++arg > nargs)
+			{
+				pfree(sinfo.data);
+				return NULL;
+			}
+		}
+	}
+
+	return sinfo.data;
+}
+
 /*
  * Returns number of rquired arguments or -1 when we cannot detect this number.
  * When no_error is true, then this function doesn't raise a error or warning.
@@ -849,29 +981,26 @@ plpgsql_check_is_sql_injection_vulnerable(PLpgSQL_checkstate *cstate,
 					case FORMAT_NPARAM_OID:
 						{
 							/* We can do check only when first argument is constant */
-							Node *first_arg = linitial(fexpr->args);
+							char	   *fmt;
+							int			loc;
 
-							if (first_arg && IsA(first_arg, Const))
+							fmt = plpgsql_check_get_const_string(cstate, linitial(fexpr->args), &loc);
+
+							if (fmt)
 							{
-								Const *c = (Const *) first_arg;
+								check_funcexpr_walker_params wp;
+								bool	is_error;
 
-								if (c->consttype == TEXTOID && !c->constisnull)
-								{
-									char *fmt = TextDatumGetCString(c->constvalue);
-									check_funcexpr_walker_params wp;
-									bool	is_error;
+								wp.cstate = cstate;
+								wp.expr = expr;
+								wp.query_str = expr->query;
 
-									wp.cstate = cstate;
-									wp.expr = expr;
-									wp.query_str = expr->query;
+								*location = -1;
+								check_fmt_string(fmt, fexpr->args, loc, &wp, &is_error, location, true);
 
-									*location = -1;
-									check_fmt_string(fmt, fexpr->args, c->location, &wp, &is_error, location, true);
-
-									/* only in this case, "format" function obviously sanitize parameters */
-									if (!is_error)
-										return *location != -1;
-								}
+								/* only in this case, "format" function obviously sanitize parameters */
+								if (!is_error)
+									return *location != -1;
 							}
 						}
 						break;
@@ -1018,6 +1147,7 @@ contain_volatile_functions_walker(Node *node, void *context)
 {
 	if (node == NULL)
 		return false;
+
 	/* Check for volatile functions in node itself */
 	if (check_functions_in_node(node, contain_volatile_functions_checker,
 								context))
@@ -1058,16 +1188,12 @@ contain_mutable_functions_walker(Node *node, void *context)
 								context))
 		return true;
 
-#if PG_VERSION_NUM < 160000
-
 	if (IsA(node, SQLValueFunction))
 	{
 		/* all variants of SQLValueFunction are stable */
 		return true;
 	}
 
-#endif
-
 	if (IsA(node, NextValueExpr))
 	{
 		/* NextValueExpr is volatile */
