@@ -42,25 +42,6 @@ PG_MODULE_MAGIC;
 PLpgSQL_plugin **plpgsql_check_plugin_var_ptr;
 PLpgSQL_plugin *prev_plpgsql_plugin;
 
-static PLpgSQL_plugin plugin_funcs = { plpgsql_check_profiler_func_init,
-									   plpgsql_check_on_func_beg,
-									   plpgsql_check_profiler_func_end,
-									   plpgsql_check_profiler_stmt_beg,
-									   plpgsql_check_profiler_stmt_end,
-									   NULL,
-									   NULL
-
-#if PG_VERSION_NUM >= 150000
-
-									  , NULL,
-									  NULL,
-									  NULL
-
-#endif
-
-									  };
-
-
 static const struct config_enum_entry plpgsql_check_mode_options[] = {
 	{"disabled", PLPGSQL_CHECK_MODE_DISABLED, false},
 	{"by_function", PLPGSQL_CHECK_MODE_BY_FUNCTION, false},
@@ -89,6 +70,14 @@ static const struct config_enum_entry tracer_level_options[] = {
 	{NULL, 0, false}
 };
 
+static const struct config_enum_entry cursors_leaks_level_options[] = {
+	{"notice", NOTICE, false},
+	{"WARNING", WARNING, false},
+	{"ERROR", ERROR, false},
+	{NULL, 0, false}
+};
+
+
 void			_PG_init(void);
 
 #if PG_VERSION_NUM < 150000
@@ -99,11 +88,11 @@ void			_PG_fini(void);
 
 #if PG_VERSION_NUM >= 150000
 
-shmem_request_hook_type prev_shmem_request_hook = NULL;
+shmem_request_hook_type plpgsql_check_prev_shmem_request_hook = NULL;
 
 #endif
 
-shmem_startup_hook_type prev_shmem_startup_hook = NULL;
+shmem_startup_hook_type plpgsql_check_prev_shmem_startup_hook = NULL;
 
 bool plpgsql_check_regress_test_mode;
 
@@ -127,7 +116,7 @@ static bool is_expected_extversion = false;
  */
 #define LOAD_EXTERNAL_FUNCTION(file, funcname)	((void *) (load_external_function(file, funcname, true, NULL)))
 
-#define EXPECTED_EXTVERSION		"2.3"
+#define EXPECTED_EXTVERSION		"2.7"
 
 void
 plpgsql_check_check_ext_version(Oid fn_oid)
@@ -141,6 +130,7 @@ plpgsql_check_check_ext_version(Oid fn_oid)
 		Assert(OidIsValid(extoid));
 
 		extver = get_extension_version(extoid);
+
 		Assert(extver);
 
 		if (strcmp(EXPECTED_EXTVERSION, extver) != 0)
@@ -206,10 +196,6 @@ _PG_init(void)
 	plpgsql_check__ns_lookup_p = (plpgsql_check__ns_lookup_t)
 		LOAD_EXTERNAL_FUNCTION("$libdir/plpgsql", "plpgsql_ns_lookup");
 
-	plpgsql_check_plugin_var_ptr = (PLpgSQL_plugin **) find_rendezvous_variable( "PLpgSQL_plugin");
-	prev_plpgsql_plugin = *plpgsql_check_plugin_var_ptr;
-	*plpgsql_check_plugin_var_ptr = &plugin_funcs;
-
 	DefineCustomBoolVariable("plpgsql_check.regress_test_mode",
 					    "reduces volatile output",
 					    NULL,
@@ -259,6 +245,14 @@ _PG_init(void)
 					    PGC_USERSET, 0,
 					    NULL, NULL, NULL);
 
+	DefineCustomBoolVariable("plpgsql_check.constants_tracing",
+					    "when is true, the variables with constant value can be used like constant",
+					    NULL,
+					    &plpgsql_check_constants_tracing,
+					    true,
+					    PGC_USERSET, 0,
+					    NULL, NULL, NULL);
+
 	DefineCustomBoolVariable("plpgsql_check.fatal_errors",
 					    "when is true, then plpgsql check stops execution on detected error",
 					    NULL,
@@ -307,6 +301,14 @@ _PG_init(void)
 					    PGC_USERSET, 0,
 					    NULL, NULL, NULL);
 
+	DefineCustomBoolVariable("plpgsql_check.tracer_show_nsubxids",
+					    "when is true, then the tracer shows number of current subxids",
+					    NULL,
+					    &plpgsql_check_tracer_show_nsubxids,
+					    false,
+					    PGC_USERSET, 0,
+					    NULL, NULL, NULL);
+
 	DefineCustomEnumVariable("plpgsql_check.tracer_verbosity",
 					    "sets the verbosity of tracer",
 					    NULL,
@@ -343,6 +345,31 @@ _PG_init(void)
 							PGC_USERSET, 0,
 							NULL, NULL, NULL);
 
+	DefineCustomEnumVariable("plpgsql_check.cursors_leaks_errlevel",
+					    "sets an error level of detection of unclosed cursors",
+					    NULL,
+					    (int *) &plpgsql_check_cursors_leaks_level,
+					    WARNING,
+					    cursors_leaks_level_options,
+					    PGC_USERSET, 0,
+					    NULL, NULL, NULL);
+
+	DefineCustomBoolVariable("plpgsql_check.cursors_leaks",
+					    "when is true, then detection of unclosed cursors is active",
+					    NULL,
+					    &plpgsql_check_cursors_leaks,
+					    true,
+					    PGC_USERSET, 0,
+					    NULL, NULL, NULL);
+
+	DefineCustomBoolVariable("plpgsql_check.strict_cursors_leaks",
+					    "when is true, then detection of unclosed cursors is executed immediately when function is finished",
+					    NULL,
+					    &plpgsql_check_cursors_leaks_strict,
+					    false,
+					    PGC_USERSET, 0,
+					    NULL, NULL, NULL);
+
 	EmitWarningsOnPlaceholders("plpgsql_check");
 
 	plpgsql_check_HashTableInit();
@@ -379,44 +406,21 @@ _PG_init(void)
 		 */
 #if PG_VERSION_NUM >= 150000
 
-		prev_shmem_request_hook = shmem_request_hook;
+		plpgsql_check_prev_shmem_request_hook = shmem_request_hook;
 		shmem_request_hook = plpgsql_check_profiler_shmem_request;
 
 #endif
 
-		prev_shmem_startup_hook = shmem_startup_hook;
+		plpgsql_check_prev_shmem_startup_hook = shmem_startup_hook;
 		shmem_startup_hook = plpgsql_check_profiler_shmem_startup;
 	}
 
-	plpgsql_check_next_needs_fmgr_hook = needs_fmgr_hook;
-	plpgsql_check_next_fmgr_hook = fmgr_hook;
-
-	needs_fmgr_hook = plpgsql_check_needs_fmgr_hook;
-	fmgr_hook = plpgsql_check_fmgr_hook;
+	plpgsql_check_init_pldbgapi2();
+	plpgsql_check_passive_check_init();
+	plpgsql_check_profiler_init();
+	plpgsql_check_tracer_init();
+	plpgsql_check_cursors_leaks_init();
 
 	inited = true;
 }
 
-#if PG_VERSION_NUM < 150000
-
-/*
- * Module unload callback
- */
-void
-_PG_fini(void)
-{
-#if PG_VERSION_NUM >= 150000
-
-	shmem_request_hook = prev_shmem_request_hook;
-
-#endif
-	shmem_startup_hook = prev_shmem_startup_hook;
-
-	/* Be more correct, and clean rendezvous variable */
-	*plpgsql_check_plugin_var_ptr = NULL;
-
-	needs_fmgr_hook = plpgsql_check_next_needs_fmgr_hook;
-	fmgr_hook = plpgsql_check_next_fmgr_hook;
-}
-
-#endif
