@@ -7,13 +7,8 @@
 #include "access/tupdesc.h"
 #include "storage/ipc.h"
 
-#if PG_VERSION_NUM >= 110000
 typedef uint64 pc_queryid;
 #define NOQUERYID				(UINT64CONST(0))
-#else
-typedef uint32 pc_queryid;
-#define NOQUERYID	(0)
-#endif
 
 #if PG_VERSION_NUM < 150000
 #define parse_analyze_fixedparams	parse_analyze
@@ -61,26 +56,40 @@ enum
 
 typedef enum
 {
-	PLPGSQL_CHECK_STMT_WALKER_PREPARE_PROFILE,
 	PLPGSQL_CHECK_STMT_WALKER_COUNT_EXEC_TIME,
 	PLPGSQL_CHECK_STMT_WALKER_PREPARE_RESULT,
 	PLPGSQL_CHECK_STMT_WALKER_COLLECT_COVERAGE
 } profiler_stmt_walker_mode;
 
+typedef enum
+{
+	PLPGSQL_CHECK_PRAGMA_ASSERT_SCHEMA,
+	PLPGSQL_CHECK_PRAGMA_ASSERT_TABLE,
+	PLPGSQL_CHECK_PRAGMA_ASSERT_COLUMN,
+} PragmaAssertType;
+
+
 typedef struct PLpgSQL_stmt_stack_item
 {
-	PLpgSQL_stmt	   *stmt;
-	char			   *label;
+	PLpgSQL_stmt *stmt;
+	char	   *label;
 	struct PLpgSQL_stmt_stack_item *outer;
-	bool				is_exception_handler;
+	bool		is_exception_handler;
+	Bitmapset  *invalidate_strconstvars;
 } PLpgSQL_stmt_stack_item;
 
+typedef struct PLpgSQL_statements
+{
+	struct PLpgSQL_statements *outer;
+	Bitmapset  *invalidate_strconstvars;
+} PLpgSQL_statements;
+
 typedef struct plpgsql_check_result_info
 {
 	int			format;						/* produced / expected format */
 	Tuplestorestate	*tuple_store;			/* target tuple store */
 	TupleDesc	tupdesc;					/* target tuple store tuple descriptor */
-	MemoryContext			query_ctx;		/* memory context for string operations */
+	MemoryContext query_ctx;				/* memory context for string operations */
 	StringInfo	sinfo;						/* buffer for multi line one value output formats */
 	bool		init_tag;					/* true, when init tag should be created */
 } plpgsql_check_result_info;
@@ -106,6 +115,7 @@ typedef struct plpgsql_check_info
 	bool		extra_warnings;
 	bool		security_warnings;
 	bool		compatibility_warnings;
+	bool		constants_tracing;
 	bool		show_profile;
 
 	bool		all_warnings;
@@ -126,17 +136,22 @@ typedef struct
 	unsigned int disable_extra_warnings : 1;
 	unsigned int disable_security_warnings : 1;
 	unsigned int disable_compatibility_warnings : 1;
+	unsigned int disable_constants_tracing : 1;
 } plpgsql_check_pragma_vector;
 
+#define CI_MAGIC		2023042922
+
 typedef struct PLpgSQL_checkstate
 {
+	int			ci_magic;
+
 	List	    *argnames;					/* function arg names */
 	char		decl_volatility;			/* declared function volatility */
 	char		volatility;					/* detected function volatility */
 	bool		has_execute_stmt;			/* detected dynamic SQL, disable volatility check */
 	bool		skip_volatility_check;		/* don't do this test for trigger */
-	PLpgSQL_execstate	   *estate;			/* check state is estate extension */
-	MemoryContext			check_cxt;
+	PLpgSQL_execstate *estate;				/* check state is estate extension */
+	MemoryContext check_cxt;
 	List	   *exprs;						/* list of all expression created by checker */
 	bool		is_active_mode;				/* true, when checking is started by plpgsql_check_function */
 	Bitmapset  *used_variables;				/* track which variables have been used; bit per varno */
@@ -144,38 +159,34 @@ typedef struct PLpgSQL_checkstate
 	PLpgSQL_stmt_stack_item *top_stmt_stack;	/* list of known labels + related command */
 	bool		found_return_query;			/* true, when code contains RETURN query */
 	bool		found_return_dyn_query;		/* true, when code contains RETURN QUERY EXECUTE */
-	Bitmapset	   *func_oids;				/* list of used (and displayed) functions */
-	Bitmapset	   *rel_oids;				/* list of used (and displayed) relations */
+	Bitmapset  *func_oids;					/* list of used (and displayed) functions */
+	Bitmapset  *rel_oids;					/* list of used (and displayed) relations */
 	bool		fake_rtd;					/* true when functions returns record */
 	plpgsql_check_result_info *result_info;
 	plpgsql_check_info *cinfo;
-	Bitmapset	   *safe_variables;			/* track which variables are safe against sql injection */
-	Bitmapset	   *out_variables;			/* what variables are used as OUT variables */
-	Bitmapset	   *protected_variables;	/* what variables should be assigned internal only */
-	Bitmapset	   *auto_variables;			/* variables initialized, used by runtime */
-	Bitmapset	   *typed_variables;		/* record variables with assigned type by pragma TYPE */
-	bool			stop_check;				/* true after error when fatal_errors option is active */
-	bool			allow_mp;				/* true, when multiple plans in plancache are allowed */
-	bool			has_mp;					/* true, when multiple plan was used */
-	bool			was_pragma;				/* true, when last expression was a plpgsql_check pragma */
+	Bitmapset  *safe_variables;				/* track which variables are safe against sql injection */
+	Bitmapset  *out_variables;				/* what variables are used as OUT variables */
+	Bitmapset  *protected_variables;		/* what variables should be assigned internal only */
+	Bitmapset  *auto_variables;				/* variables initialized, used by runtime */
+	Bitmapset  *typed_variables;			/* record variables with assigned type by pragma TYPE */
+	bool		stop_check;					/* true after error when fatal_errors option is active */
+	bool		allow_mp;					/* true, when multiple plans in plancache are allowed */
+	bool		has_mp;						/* true, when multiple plan was used */
+	bool		was_pragma;					/* true, when last expression was a plpgsql_check pragma */
 	plpgsql_check_pragma_vector pragma_vector;
 	Oid			pragma_foid;				/* oid of plpgsql_check pragma function */
+	char	  **strconstvars;				/* the values of string variables where the value is constant */
+	PLpgSQL_statements *top_stmts;			/* pointer to current statement group */
 } PLpgSQL_checkstate;
 
 typedef struct
 {
-	int		statements;
-	int		branches;
-	int		executed_statements;
-	int		executed_branches;
+	int			statements;
+	int			branches;
+	int			executed_statements;
+	int			executed_branches;
 } coverage_state;
 
-typedef struct
-{
-	unsigned long int run_id;
-	int		level;
-} tracer_info;
-
 /*
  * function from assign.c
  */
@@ -245,6 +256,7 @@ extern bool plpgsql_check_extra_warnings;
 extern bool plpgsql_check_performance_warnings;
 extern bool plpgsql_check_compatibility_warnings;
 extern bool plpgsql_check_fatal_errors;
+extern bool plpgsql_check_constants_tracing;
 extern int plpgsql_check_mode;
 
 /*
@@ -258,11 +270,15 @@ extern bool plpgsql_check_is_sql_injection_vulnerable(PLpgSQL_checkstate *cstate
 extern bool plpgsql_check_contain_volatile_functions(Node *clause, PLpgSQL_checkstate *cstate);
 extern bool plpgsql_check_contain_mutable_functions(Node *clause, PLpgSQL_checkstate *cstate);
 extern bool plpgsql_check_vardno_is_used_for_reading(Node *node, int dno);
+extern char *plpgsql_check_get_formatted_string(PLpgSQL_checkstate *cstate, const char *fmt, List *args,
+	bool *found_ident_placeholder, bool *found_literal_placeholder, bool *expr_is_const);
 
 /*
  * functions from check_expr.c
  */
-extern char *plpgsql_check_expr_get_string(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool *isnull);
+extern char *plpgsql_check_expr_get_string(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, int *location);
+extern char *plpgsql_check_get_tracked_const(PLpgSQL_checkstate *cstate, Node *node);
+extern char *plpgsql_check_get_const_string(PLpgSQL_checkstate *cstate, Node *node, int *location);
 extern void plpgsql_check_expr_with_scalar_type(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, Oid expected_typoid, bool required);
 extern void plpgsql_check_returned_expr(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool is_expression);
 extern void plpgsql_check_expr_as_rvalue(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr,
@@ -280,20 +296,16 @@ extern void plpgsql_check_expr_generic_with_parser_setup(PLpgSQL_checkstate *cst
 	ParserSetupHook parser_setup, void *arg);
 
 extern Node *plpgsql_check_expr_get_node(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool force_plan_checks);
-extern char *plpgsql_check_const_to_string(Const *c);
+extern char *plpgsql_check_const_to_string(Node *node, int *location);
 extern CachedPlanSource *plpgsql_check_get_plan_source(PLpgSQL_checkstate *cstate, SPIPlanPtr plan);
 
-#if PG_VERSION_NUM >= 110000
-
 extern void plpgsql_check_assignment_to_variable(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr,
 	PLpgSQL_variable *targetvar, int targetdno);
 
-#endif
-
 /*
  * functions from report.c
  */
-extern char * plpgsql_check_datum_get_refname(PLpgSQL_datum *d);
+extern char * plpgsql_check_datum_get_refname(PLpgSQL_checkstate *cstate, PLpgSQL_datum *d);
 extern void plpgsql_check_report_unused_variables(PLpgSQL_checkstate *cstate);
 extern void plpgsql_check_report_too_high_volatility(PLpgSQL_checkstate *cstate);
 extern bool is_internal_variable(PLpgSQL_checkstate *cstate, PLpgSQL_variable *var);
@@ -309,15 +321,9 @@ extern void plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, i
  */
 extern TupleDesc plpgsql_check_expr_get_desc(PLpgSQL_checkstate *cstate, PLpgSQL_expr *query,
 	bool use_element_type, bool expand_record, bool is_expression, Oid *first_level_typoid);
-
 extern void plpgsql_check_recvar_info(PLpgSQL_rec *rec, Oid *typoid, int32 *typmod);
-
-#if PG_VERSION_NUM >= 110000
-
 extern PLpgSQL_row * plpgsql_check_CallExprGetRowTarget(PLpgSQL_checkstate *cstate, PLpgSQL_expr *CallExpr);
 
-#endif
-
 /*
  * functions from parser.c
  */
@@ -325,6 +331,8 @@ extern Oid plpgsql_check_parse_name_or_signature(char *name_or_signature);
 extern bool plpgsql_check_pragma_type(PLpgSQL_checkstate *cstate, const char *str, PLpgSQL_nsitem *ns, int lineno);
 extern bool plpgsql_check_pragma_table(PLpgSQL_checkstate *cstate, const char *str, int lineno);
 extern bool plpgsql_check_pragma_sequence(PLpgSQL_checkstate *cstate, const char *str, int lineno);
+extern bool plpgsql_check_pragma_assert(PLpgSQL_checkstate *cstate, PragmaAssertType pat, const char *str,
+										PLpgSQL_nsitem *ns, int lineno);
 extern void plpgsql_check_search_comment_options(plpgsql_check_info *cinfo);
 extern char *plpgsql_check_process_echo_string(char *str, plpgsql_check_info *cinfo);
 
@@ -345,12 +353,6 @@ extern void plpgsql_check_profiler_shmem_startup(void);
 extern Size plpgsql_check_shmem_size(void);
 extern void plpgsql_check_profiler_init_hash_tables(void);
 
-extern void plpgsql_check_profiler_func_init(PLpgSQL_execstate *estate, PLpgSQL_function *func);
-extern void plpgsql_check_profiler_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func);
-extern void plpgsql_check_profiler_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *func);
-extern void plpgsql_check_profiler_stmt_beg(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
-extern void plpgsql_check_profiler_stmt_end(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
-
 extern void plpgsql_check_iterate_over_profile(plpgsql_check_info *cinfo, profiler_stmt_walker_mode mode,
    plpgsql_check_result_info *ri, coverage_state *cs);
 
@@ -360,15 +362,10 @@ extern void plpgsql_check_profiler_iterate_over_all_profiles(plpgsql_check_resul
 extern void plpgsql_check_init_trace_info(PLpgSQL_execstate *estate);
 extern bool plpgsql_check_get_trace_info(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt, PLpgSQL_execstate **outer_estate, int *frame_num, int *level, instr_time *start_time);
 
-extern bool plpgsql_check_needs_fmgr_hook(Oid fn_oid);
-extern void plpgsql_check_fmgr_hook(FmgrHookEventType event, FmgrInfo *flinfo, Datum *private);
-
-#if PG_VERSION_NUM >= 120000
-
 extern void plpgsql_check_get_trace_stmt_info(PLpgSQL_execstate *estate, int stmt_id, instr_time **start_time);
 extern bool *plpgsql_check_get_disable_tracer_on_stack(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
 
-#endif
+extern void plpgsql_check_profiler_init(void);
 
 /*
  * functions and variables from tracer.c
@@ -385,57 +382,114 @@ extern PGErrorVerbosity plpgsql_check_tracer_verbosity;
 extern PGErrorVerbosity plpgsql_check_trace_assert_verbosity;
 
 extern bool plpgsql_check_regress_test_mode;
+extern bool plpgsql_check_tracer_show_nsubxids;
 
-extern void plpgsql_check_tracer_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func);
-extern void plpgsql_check_tracer_on_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *func);
-extern void plpgsql_check_tracer_on_stmt_beg(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
-extern void plpgsql_check_tracer_on_stmt_end(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
 extern void plpgsql_check_trace_assert_on_stmt_beg(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
 
 extern void plpgsql_check_set_stmt_group_number(PLpgSQL_stmt *stmt, int *group_numbers, int *parent_group_numbers, int sgn, int *cgn, int psgn);
 
+extern void plpgsql_check_tracer_init(void);
+
 /*
  * variables from pragma.c
  */
 extern void plpgsql_check_pragma_apply(PLpgSQL_checkstate *cstate, char *pragma_str, PLpgSQL_nsitem *ns, int lineno);
 
-extern plpgsql_check_pragma_vector plpgsql_check_runtime_pragma_vector;
-extern bool plpgsql_check_runtime_pragma_vector_changed;
+/*
+ * pldbgapi2 statement plugin2 info. This info is created
+ * when function is first started and it is cached in session.
+ */
+typedef struct plpgsql_check_plugin2_stmt_info
+{
+	int			level;
+	int			natural_id;
+	int			parent_id;
+	const char *typname;
+	bool		is_invisible;
+	bool		is_container;
+} plpgsql_check_plugin2_stmt_info;
+
+/*
+ * functions from pldbgapi2
+ */
+typedef struct plpgsql_check_plugin2
+{
+	/* Function pointers set up by the plugin */
+	void		(*func_setup2) (PLpgSQL_execstate *estate, PLpgSQL_function *func, void **plugin2_info);
+	void		(*func_beg2) (PLpgSQL_execstate *estate, PLpgSQL_function *func, void **plugin2_info);
+	void		(*func_end2) (PLpgSQL_execstate *estate, PLpgSQL_function *func, void **plugin2_info);
+	void		(*func_end2_aborted) (Oid fn_oid, void **plugin2_info);
+	void		(*stmt_beg2) (PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt, void **plugin2_info);
+	void		(*stmt_end2) (PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt, void **plugin2_info);
+	void		(*stmt_end2_aborted) (Oid fn_oid, int stmtid, void **plugin2_info);
+
+	/* Function pointers set by PL/pgSQL itself */
+	void		(*error_callback) (void *arg);
+	void		(*assign_expr) (PLpgSQL_execstate *estate,
+								PLpgSQL_datum *target,
+								PLpgSQL_expr *expr);
+
+	void		(*assign_value) (PLpgSQL_execstate *estate,
+								 PLpgSQL_datum *target,
+								 Datum value, bool isNull,
+								 Oid valtype, int32 valtypmod);
+	void		(*eval_datum) (PLpgSQL_execstate *estate, PLpgSQL_datum *datum,
+							   Oid *typeId, int32 *typetypmod,
+							   Datum *value, bool *isnull);
+	Datum		(*cast_value) (PLpgSQL_execstate *estate,
+							   Datum value, bool *isnull,
+							   Oid valtype, int32 valtypmod,
+							   Oid reqtype, int32 reqtypmod);
+} plpgsql_check_plugin2;
+
+extern void plpgsql_check_register_pldbgapi2_plugin(plpgsql_check_plugin2 *plugin2);
+extern void plpgsql_check_init_pldbgapi2(void);
+
+extern plpgsql_check_plugin2_stmt_info *plpgsql_check_get_current_stmt_info(int stmtid);
+
+extern plpgsql_check_plugin2_stmt_info *plpgsql_check_get_current_stmts_info(void);
+extern plpgsql_check_plugin2_stmt_info *plpgsql_check_get_stmts_info(PLpgSQL_function *func);
+
+extern int *plpgsql_check_get_current_stmtid_map(void);
+extern int *plpgsql_check_get_stmtid_map(PLpgSQL_function *func);
+
+extern char *plpgsql_check_get_current_func_info_name(void);
+extern char *plpgsql_check_get_current_func_info_signature(void);
+
+#if PG_VERSION_NUM < 150000
+
+extern void plpgsql_check_finish_pldbgapi2(void);
+
+#endif
+
+/*
+ * functions from cursors_leaks.c
+ */
+extern bool plpgsql_check_cursors_leaks;
+extern bool plpgsql_check_cursors_leaks_strict;
+extern int plpgsql_check_cursors_leaks_level;
+
+extern void plpgsql_check_cursors_leaks_init(void);
 
 /*
  * functions from plpgsql_check.c
  */
 
 #if PG_VERSION_NUM >= 150000
-extern shmem_request_hook_type prev_shmem_request_hook;
+extern shmem_request_hook_type plpgsql_check_prev_shmem_request_hook;
 #endif
-extern shmem_startup_hook_type prev_shmem_startup_hook;
+extern shmem_startup_hook_type plpgsql_check_prev_shmem_startup_hook;
 
 extern PLpgSQL_plugin **plpgsql_check_plugin_var_ptr;
-extern PLpgSQL_plugin *prev_plpgsql_plugin;
 
 extern void plpgsql_check_check_ext_version(Oid fn_oid);
-
-#if PG_VERSION_NUM > 110005
-
-#define PLPGSQL_BUILD_DATATYPE_4		1
-
-#endif
+extern void plpgsql_check_passive_check_init(void);
 
 /*
  * Links to function in plpgsql module
  */
-
-#ifdef PLPGSQL_BUILD_DATATYPE_4
-
 typedef PLpgSQL_type *(*plpgsql_check__build_datatype_t) (Oid typeOid, int32 typmod, Oid collation, TypeName *origtypname);
 
-#else
-
-typedef PLpgSQL_type *(*plpgsql_check__build_datatype_t) (Oid typeOid, int32 typmod, Oid collation);
-
-#endif
-
 extern plpgsql_check__build_datatype_t plpgsql_check__build_datatype_p;
 typedef PLpgSQL_function *(*plpgsql_check__compile_t) (FunctionCallInfo fcinfo, bool forValidator);
 extern plpgsql_check__compile_t plpgsql_check__compile_p;
@@ -478,28 +532,9 @@ extern plpgsql_check__ns_lookup_t plpgsql_check__ns_lookup_p;
 #define TupleDescAttr(tupdesc, i) ((tupdesc)->attrs[(i)])
 #endif
 
-#if PG_VERSION_NUM >= 110000
-
 #define recvar_tuple(rec)		(rec->erh ? expanded_record_get_tuple(rec->erh) : NULL)
 #define recvar_tupdesc(rec)		(rec->erh ? expanded_record_fetch_tupdesc(rec->erh) : NULL)
 
-#else
-
-#define recvar_tuple(rec)		(rec->tup)
-#define recvar_tupdesc(rec)		(rec->tupdesc)
-
-#endif
-
-#if PG_VERSION_NUM >= 100000
-
-#define PLPGSQL_STMT_TYPES
-
-#else
-
-#define PLPGSQL_STMT_TYPES		(enum PLpgSQL_stmt_types)
-
-#endif
-
 #define FUNCS_PER_USER		128 /* initial table size */
 
 #ifdef _MSC_VER
