@@ -28,64 +28,42 @@
 
 PG_FUNCTION_INFO_V1(plpgsql_check_pragma);
 
-plpgsql_check_pragma_vector plpgsql_check_runtime_pragma_vector;
-
-bool plpgsql_check_runtime_pragma_vector_changed = false;
-
 static void
-runtime_pragma_apply(plpgsql_check_pragma_vector *pv,
-					 char *pragma_str)
+runtime_pragma_apply(char *pragma_str)
 {
-	while (*pragma_str == ' ')
+	while (scanner_isspace(*pragma_str))
 		pragma_str++;
 
 	if (strncasecmp(pragma_str, "STATUS:", 7) == 0)
 	{
 		pragma_str += 7;
 
-		while (*pragma_str == ' ')
+		while (scanner_isspace(*pragma_str))
 			pragma_str++;
 
 		if (strcasecmp(pragma_str, "TRACER") == 0)
 			elog(NOTICE, "tracer is %s",
-					pv->disable_tracer ? "disabled" : "enabled");
+					plpgsql_check_tracer ? "enabled" : "disabled");
 	}
 	else if (strncasecmp(pragma_str, "ENABLE:", 7) == 0)
 	{
 		pragma_str += 7;
 
-		while (*pragma_str == ' ')
+		while (scanner_isspace(*pragma_str))
 			pragma_str++;
 
 		if (strcasecmp(pragma_str, "TRACER") == 0)
-		{
-			pv->disable_tracer = false;
-
-#if PG_VERSION_NUM < 120000
-
-			elog(WARNING, "pragma ENABLE:TRACER is ignored on PostgreSQL 11 and older");
-
-#endif
-
-		}
+			plpgsql_check_tracer = true;
 	}
 	else if (strncasecmp(pragma_str, "DISABLE:", 8) == 0)
 	{
 		pragma_str += 8;
 
-		while (*pragma_str == ' ')
+		while (scanner_isspace(*pragma_str))
 			pragma_str++;
 
 		if (strcasecmp(pragma_str, "TRACER") == 0)
-		{
-			pv->disable_tracer = true;
-
-#if PG_VERSION_NUM < 120000
-
-			elog(WARNING, "pragma DISABLE:TRACER is ignored on PostgreSQL 11 and older");
-
-#endif
-		}
+			plpgsql_check_tracer = false;
 	}
 }
 
@@ -100,7 +78,7 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 
 	Assert(cstate);
 
-	while (*pragma_str == ' ')
+	while (scanner_isspace(*pragma_str))
 		pragma_str++;
 
 	if (strncasecmp(pragma_str, "ECHO:", 5) == 0)
@@ -111,7 +89,7 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 	{
 		pragma_str += 7;
 
-		while (*pragma_str == ' ')
+		while (scanner_isspace(*pragma_str))
 			pragma_str++;
 
 		if (strcasecmp(pragma_str, "CHECK") == 0)
@@ -135,6 +113,9 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 		else if (strcasecmp(pragma_str, "COMPATIBILITY_WARNINGS") == 0)
 			elog(NOTICE, "compatibility_warnings is %s",
 					pv->disable_compatibility_warnings ? "disabled" : "enabled");
+		else if (strcasecmp(pragma_str, "CONSTANTS_TRANCING") == 0)
+			elog(NOTICE, "constants_traising is %s",
+					pv->disable_constants_tracing ? "disabled" : "enabled");
 		else
 		{
 			elog(WARNING, "unsuported pragma: %s", pragma_str);
@@ -145,22 +126,13 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 	{
 		pragma_str += 7;
 
-		while (*pragma_str == ' ')
+		while (scanner_isspace(*pragma_str))
 			pragma_str++;
 
 		if (strcasecmp(pragma_str, "CHECK") == 0)
 			pv->disable_check = false;
 		else if (strcasecmp(pragma_str, "TRACER") == 0)
-		{
 			pv->disable_tracer = false;
-
-#if PG_VERSION_NUM < 120000
-
-			elog(WARNING, "pragma ENABLE:TRACER is ignored on PostgreSQL 11 and older");
-
-#endif
-
-		}
 		else if (strcasecmp(pragma_str, "OTHER_WARNINGS") == 0)
 			pv->disable_other_warnings = false;
 		else if (strcasecmp(pragma_str, "PERFORMANCE_WARNINGS") == 0)
@@ -171,6 +143,8 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 			pv->disable_security_warnings = false;
 		else if (strcasecmp(pragma_str, "COMPATIBILITY_WARNINGS") == 0)
 			pv->disable_compatibility_warnings = false;
+		else if (strcasecmp(pragma_str, "CONSTANTS_TRACING") == 0)
+			pv->disable_constants_tracing = false;
 		else
 		{
 			elog(WARNING, "unsuported pragma: %s", pragma_str);
@@ -181,21 +155,13 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 	{
 		pragma_str += 8;
 
-		while (*pragma_str == ' ')
+		while (scanner_isspace(*pragma_str))
 			pragma_str++;
 
 		if (strcasecmp(pragma_str, "CHECK") == 0)
 			pv->disable_check = true;
 		else if (strcasecmp(pragma_str, "TRACER") == 0)
-		{
 			pv->disable_tracer = true;
-
-#if PG_VERSION_NUM < 120000
-
-			elog(WARNING, "pragma DISABLE:TRACER is ignored on PostgreSQL 11 and older");
-
-#endif
-		}
 		else if (strcasecmp(pragma_str, "OTHER_WARNINGS") == 0)
 			pv->disable_other_warnings = true;
 		else if (strcasecmp(pragma_str, "PERFORMANCE_WARNINGS") == 0)
@@ -206,6 +172,8 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 			pv->disable_security_warnings = true;
 		else if (strcasecmp(pragma_str, "COMPATIBILITY_WARNINGS") == 0)
 			pv->disable_compatibility_warnings = true;
+		else if (strcasecmp(pragma_str, "CONSTANTS_TRACING") == 0)
+			pv->disable_constants_tracing = true;
 		else
 			elog(WARNING, "unsuported pragma: %s", pragma_str);
 	}
@@ -221,6 +189,24 @@ pragma_apply(PLpgSQL_checkstate *cstate,
 	{
 		is_valid = plpgsql_check_pragma_sequence(cstate, pragma_str + 9, lineno);
 	}
+	else if (strncasecmp(pragma_str, "ASSERT-SCHEMA:", 14) == 0)
+	{
+		is_valid = plpgsql_check_pragma_assert(cstate,
+											   PLPGSQL_CHECK_PRAGMA_ASSERT_SCHEMA,
+											   pragma_str + 14, ns, lineno);
+	}
+	else if (strncasecmp(pragma_str, "ASSERT-TABLE:", 13) == 0)
+	{
+		is_valid = plpgsql_check_pragma_assert(cstate,
+											   PLPGSQL_CHECK_PRAGMA_ASSERT_TABLE,
+											   pragma_str + 13, ns, lineno);
+	}
+	else if (strncasecmp(pragma_str, "ASSERT-COLUMN:", 14) == 0)
+	{
+		is_valid = plpgsql_check_pragma_assert(cstate,
+											   PLPGSQL_CHECK_PRAGMA_ASSERT_COLUMN,
+											   pragma_str + 14, ns, lineno);
+	}
 	else
 	{
 		elog(WARNING, "unsupported pragma: %s", pragma_str);
@@ -262,10 +248,7 @@ plpgsql_check_pragma(PG_FUNCTION_ARGS)
 
 		pragma_str = TextDatumGetCString(value);
 
-		runtime_pragma_apply(&plpgsql_check_runtime_pragma_vector, pragma_str);
-
-		plpgsql_check_runtime_pragma_vector_changed = true;
-
+		runtime_pragma_apply(pragma_str);
 		pfree(pragma_str);
 	}
 
