@@ -32,25 +32,9 @@ static int possibly_closed(int c);
 static int merge_closing(int c, int c_local, List **exceptions, List *exceptions_local, int err_code);
 static bool exception_matches_conditions(int sqlerrstate, PLpgSQL_condition *cond);
 static bool found_shadowed_variable(char *varname, PLpgSQL_stmt_stack_item *current, PLpgSQL_checkstate *cstate);
-
-#if PG_VERSION_NUM >= 110000
-
 static void check_dynamic_sql(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, PLpgSQL_expr *query, bool into, PLpgSQL_variable *target, List *params);
-
-#else
-
-static void check_dynamic_sql(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, PLpgSQL_expr *query, bool into, PLpgSQL_row *row, PLpgSQL_rec *rec, List *params);
-
-#endif
-
-#if PG_VERSION_NUM >= 110000
-
 static bool is_inside_protected_block(PLpgSQL_stmt_stack_item *current);
 
-#endif
-
-#if PG_VERSION_NUM >= 110000
-
 static void
 check_variable(PLpgSQL_checkstate *cstate, PLpgSQL_variable *var)
 {
@@ -92,24 +76,11 @@ check_variable(PLpgSQL_checkstate *cstate, PLpgSQL_variable *var)
 	elog(ERROR, "unsupported dtype %d", var->dtype);
 }
 
-#endif
-
 bool
 plpgsql_check_is_reserved_keyword(char *name)
 {
 	int		i;
 
-#if PG_VERSION_NUM < 120000
-
-	for (i = 0; i < NumScanKeywords; i++)
-	{
-		if (ScanKeywords[i].category == RESERVED_KEYWORD &&
-				strcmp(name, ScanKeywords[i].name) == 0)
-			return true;
-	}
-
-#else
-
 	for (i = 0; i < ScanKeywords.num_keywords; i++)
 	{
 		if (ScanKeywordCategories[i] == RESERVED_KEYWORD)
@@ -122,11 +93,90 @@ plpgsql_check_is_reserved_keyword(char *name)
 		}
 	}
 
-#endif
-
 	return false;
 }
 
+/*
+ * does warning checks - variable shadowing, function's argument shadowing and
+ * using keywords as variable's name
+ */
+static void
+check_variable_name(PLpgSQL_checkstate *cstate,
+					PLpgSQL_stmt_stack_item *outer_stmt_stack,
+					int dno)
+{
+	char	   *refname;
+	PLpgSQL_datum *d = cstate->estate->func->datums[dno];
+
+	refname = plpgsql_check_datum_get_refname(cstate, d);
+	if (refname != NULL)
+	{
+		ListCell   *l;
+		bool		is_auto = bms_is_member(d->dno, cstate->auto_variables);
+
+		if (plpgsql_check_is_reserved_keyword(refname))
+		{
+			StringInfoData str;
+
+			initStringInfo(&str);
+			appendStringInfo(&str, "name of variable \"%s\" is reserved keyword",
+							 refname);
+
+			plpgsql_check_put_error(cstate,
+						  0, 0,
+						  str.data,
+						  "The reserved keyword was used as variable name.",
+						  NULL,
+						  PLPGSQL_CHECK_WARNING_OTHERS,
+						  0, NULL, NULL);
+			pfree(str.data);
+		}
+
+		foreach(l, cstate->argnames)
+		{
+			char	   *argname = (char *) lfirst(l);
+
+			if (strcmp(argname, refname) == 0)
+			{
+				StringInfoData str;
+
+				initStringInfo(&str);
+				appendStringInfo(&str, "parameter \"%s\" is overlapped",
+								 refname);
+
+				plpgsql_check_put_error(cstate,
+							  0, 0,
+							  str.data,
+							  is_auto ?
+							  "Local auto variable overlap function parameter." :
+							  "Local variable overlap function parameter.",
+							  NULL,
+							  PLPGSQL_CHECK_WARNING_OTHERS,
+							  0, NULL, NULL);
+				pfree(str.data);
+			}
+		}
+
+		if (found_shadowed_variable(refname, outer_stmt_stack, cstate))
+		{
+			StringInfoData str;
+
+			initStringInfo(&str);
+			appendStringInfo(&str, "%svariable \"%s\" shadows a previously defined variable",
+								 is_auto ? "auto " : "", refname);
+
+			plpgsql_check_put_error(cstate,
+							  0, 0,
+							  str.data,
+							  NULL,
+							  "SET plpgsql.extra_warnings TO 'shadowed_variables'",
+							  PLPGSQL_CHECK_WARNING_EXTRA,
+							  0, NULL, NULL);
+			pfree(str.data);
+		}
+	}
+}
+
 /*
  * walk over all plpgsql statements - search and check expressions
  *
@@ -138,7 +188,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 	PLpgSQL_function *func;
 	ResourceOwner oldowner;
 	MemoryContext oldCxt = CurrentMemoryContext;
-	PLpgSQL_stmt_stack_item *outer_stmt;
+	PLpgSQL_stmt_stack_item *outer_stmt_stack;
 	plpgsql_check_pragma_vector pragma_vector;
 
 	if (stmt == NULL)
@@ -156,7 +206,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 	/*
 	 * Attention - returns NULL, when there are not any outer level
 	 */
-	outer_stmt = push_stmt_to_stmt_stack(cstate);
+	outer_stmt_stack = push_stmt_to_stmt_stack(cstate);
 
 	oldowner = CurrentResourceOwner;
 	BeginInternalSubTransaction(NULL);
@@ -164,19 +214,16 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 	PG_TRY();
 	{
-		switch (PLPGSQL_STMT_TYPES stmt->cmd_type)
+		switch (stmt->cmd_type)
 		{
 			case PLPGSQL_STMT_BLOCK:
 				{
 					PLpgSQL_stmt_block *stmt_block = (PLpgSQL_stmt_block *) stmt;
 					int			i;
-					PLpgSQL_datum *d;
 
 					for (i = 0; i < stmt_block->n_initvars; i++)
 					{
-						char	   *refname;
-
-						d = func->datums[stmt_block->initvarnos[i]];
+						PLpgSQL_datum *d = func->datums[stmt_block->initvarnos[i]];
 
 						if (d->dtype == PLPGSQL_DTYPE_VAR ||
 							d->dtype == PLPGSQL_DTYPE_ROW ||
@@ -192,8 +239,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 							cstate->estate->err_text = str.data;
 
-#if PG_VERSION_NUM >= 110000
-
 							if (var->default_val)
 								plpgsql_check_assignment(cstate,
 														 var->default_val,
@@ -201,89 +246,15 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 														 NULL,
 														 var->dno);
 
-#else
-
-							if (d->dtype == PLPGSQL_DTYPE_VAR &&
-									((PLpgSQL_var *) var)->default_val)
-								plpgsql_check_assignment(cstate,
-														 ((PLpgSQL_var *) var)->default_val,
-														 NULL,
-														 NULL,
-														 var->dno);
-
-
-#endif
-
 							cstate->estate->err_text = NULL;
 							pfree(str.data);
 						}
 
-						refname = plpgsql_check_datum_get_refname(d);
-						if (refname != NULL)
-						{
-							ListCell   *l;
-
-							if (plpgsql_check_is_reserved_keyword(refname))
-							{
-								StringInfoData str;
-
-								initStringInfo(&str);
-								appendStringInfo(&str, "name of variable \"%s\" is reserved keyword",
-												 refname);
-
-								plpgsql_check_put_error(cstate,
-											  0, 0,
-											  str.data,
-											  "The reserved keyword was used as variable name.",
-											  NULL,
-											  PLPGSQL_CHECK_WARNING_OTHERS,
-											  0, NULL, NULL);
-								pfree(str.data);
-							}
-
-							foreach(l, cstate->argnames)
-							{
-								char	   *argname = (char *) lfirst(l);
-
-								if (strcmp(argname, refname) == 0)
-								{
-									StringInfoData str;
-
-									initStringInfo(&str);
-									appendStringInfo(&str, "parameter \"%s\" is overlapped",
-													 refname);
-
-									plpgsql_check_put_error(cstate,
-												  0, 0,
-												  str.data,
-												  "Local variable overlap function parameter.",
-												  NULL,
-												  PLPGSQL_CHECK_WARNING_OTHERS,
-												  0, NULL, NULL);
-									pfree(str.data);
-								}
-							}
-
-							if (found_shadowed_variable(refname, outer_stmt, cstate))
-							{
-								StringInfoData str;
-
-								initStringInfo(&str);
-								appendStringInfo(&str, "variable \"%s\" shadows a previously defined variable",
-													 refname);
-
-								plpgsql_check_put_error(cstate,
-												  0, 0,
-												  str.data,
-												  NULL,
-												  "SET plpgsql.extra_warnings TO 'shadowed_variables'",
-												  PLPGSQL_CHECK_WARNING_EXTRA,
-												  0, NULL, NULL);
-								pfree(str.data);
-							}
-						}
+						check_variable_name(cstate, outer_stmt_stack, d->dno);
 					}
 
+					check_variable_name(cstate, outer_stmt_stack, cstate->estate->found_varno);
+
 					check_stmts(cstate, stmt_block->body, closing, exceptions);
 
 					if (stmt_block->exceptions)
@@ -519,7 +490,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 			case PLPGSQL_STMT_CASE:
 				{
 					PLpgSQL_stmt_case *stmt_case = (PLpgSQL_stmt_case *) stmt;
-					Oid			result_oid;
 					int			closing_local;
 					List	    *exceptions_local;
 					ListCell    *l;
@@ -528,6 +498,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					if (stmt_case->t_expr != NULL)
 					{
 						PLpgSQL_var *t_var = (PLpgSQL_var *) cstate->estate->datums[stmt_case->t_varno];
+						Oid			result_oid;
 
 						/*
 						 * we need to set hidden variable type
@@ -554,21 +525,11 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 						 */
 						if (t_var->datatype->typoid != result_oid)
 
-#ifdef PLPGSQL_BUILD_DATATYPE_4
-
 							t_var->datatype = plpgsql_check__build_datatype_p(result_oid,
 																	 -1,
 								   cstate->estate->func->fn_input_collation,
 								   t_var->datatype->origtypname);
 
-#else
-
-							t_var->datatype = plpgsql_check__build_datatype_p(result_oid,
-																	 -1,
-								   cstate->estate->func->fn_input_collation);
-
-#endif
-
 						ReleaseTupleDesc(tupdesc);
 					}
 
@@ -642,6 +603,8 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					cstate->protected_variables = bms_add_member(cstate->protected_variables, dno);
 					cstate->auto_variables = bms_add_member(cstate->auto_variables, dno);
 
+					check_variable_name(cstate, outer_stmt_stack, dno);
+
 					check_stmts(cstate, stmt_fori->body, &closing_local, &exceptions_local);
 					*closing = possibly_closed(closing_local);
 				}
@@ -653,24 +616,12 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					int		closing_local;
 					List   *exceptions_local;
 
-#if PG_VERSION_NUM >= 110000
-
 					check_variable(cstate, stmt_fors->var);
 
 					/* we need to set hidden variable type */
 					plpgsql_check_assignment_to_variable(cstate, stmt_fors->query,
 									 stmt_fors->var, -1);
 
-#else
-
-					plpgsql_check_row_or_rec(cstate, stmt_fors->row, stmt_fors->rec);
-
-					/* we need to set hidden variable type */
-					plpgsql_check_assignment(cstate, stmt_fors->query,
-									 stmt_fors->rec, stmt_fors->row, -1);
-
-#endif
-
 					check_stmts(cstate, stmt_fors->body, &closing_local, &exceptions_local);
 					*closing = possibly_closed(closing_local);
 				}
@@ -683,32 +634,13 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					int		closing_local;
 					List   *exceptions_local;
 
-#if PG_VERSION_NUM >= 110000
-
 					check_variable(cstate, stmt_forc->var);
-
-#else
-
-					plpgsql_check_row_or_rec(cstate, stmt_forc->row, stmt_forc->rec);
-
-#endif
-
 					plpgsql_check_expr_as_sqlstmt_data(cstate, stmt_forc->argquery);
 
-#if PG_VERSION_NUM >= 110000
-
 					if (var->cursor_explicit_expr != NULL)
 						plpgsql_check_assignment_to_variable(cstate, var->cursor_explicit_expr,
 										 stmt_forc->var, -1);
 
-#else
-
-					if (var->cursor_explicit_expr != NULL)
-						plpgsql_check_assignment(cstate, var->cursor_explicit_expr,
-										 stmt_forc->rec, stmt_forc->row, -1);
-
-#endif
-
 					check_stmts(cstate, stmt_forc->body, &closing_local, &exceptions_local);
 					*closing = possibly_closed(closing_local);
 
@@ -727,18 +659,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 									  stmt,
 									  stmt_dynfors->query,
 									  true,
-
-#if PG_VERSION_NUM >= 110000
-
 									  stmt_dynfors->var,
-
-#else
-
-									  stmt_dynfors->row,
-									  stmt_dynfors->rec,
-
-#endif
-
 									  stmt_dynfors->params);
 
 					check_stmts(cstate, stmt_dynfors->body, &closing_local, &exceptions_local);
@@ -784,7 +705,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					if (stmt_exit->label != NULL)
 					{
 						PLpgSQL_stmt *labeled_stmt = find_stmt_with_label(stmt_exit->label,
-												    outer_stmt);
+												    outer_stmt_stack);
 						if (labeled_stmt == NULL)
 							ereport(ERROR,
 								(errcode(ERRCODE_SYNTAX_ERROR),
@@ -799,7 +720,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					}
 					else
 					{
-						if (find_nearest_loop(outer_stmt) == NULL)
+						if (find_nearest_loop(outer_stmt_stack) == NULL)
 							ereport(ERROR,
 								(errcode(ERRCODE_SYNTAX_ERROR),
 								 errmsg("%s cannot be used outside a loop",
@@ -941,16 +862,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 									(errcode(ERRCODE_SYNTAX_ERROR),
 									 errmsg("cannot use RETURN NEXT in a non-SETOF function")));
 
-#if PG_VERSION_NUM >= 110000
-
 						tupdesc = estate->tuple_store_desc;
-
-#else
-
-						tupdesc = estate->rettupdesc;
-
-#endif
-
 						natts = tupdesc ? tupdesc->natts : 0;
 
 						switch (retvar->dtype)
@@ -973,7 +885,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 							case PLPGSQL_DTYPE_REC:
 								{
 									PLpgSQL_rec *rec = (PLpgSQL_rec *) retvar;
-									TupleConversionMap *tupmap;
 
 									if (!HeapTupleIsValid(recvar_tuple(rec)))
 										ereport(ERROR,
@@ -985,6 +896,8 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 									if (tupdesc)
 									{
+										TupleConversionMap *tupmap;
+
 										tupmap = convert_tuples_by_position(recvar_tupdesc(rec),
 																tupdesc,
 											gettext_noop("wrong record type supplied in RETURN NEXT"));
@@ -996,16 +909,14 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 							case PLPGSQL_DTYPE_ROW:
 								{
-									PLpgSQL_row *row = (PLpgSQL_row *) retvar;
-									bool	row_is_valid_result;
-
-									row_is_valid_result = true;
-
 									if (tupdesc)
 									{
+										bool		row_is_valid_result = true;
+										PLpgSQL_row *row = (PLpgSQL_row *) retvar;
+
 										if (row->nfields == natts)
 										{
-											int		i;
+											int			i;
 
 											for (i = 0; i < natts; i++)
 											{
@@ -1061,17 +972,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 										  stmt,
 										  stmt_rq->dynquery,
 										  false,
-
-#if PG_VERSION_NUM >= 110000
-
 										  NULL,
-
-#else
-
-										  NULL, NULL,
-
-#endif
-
 										  stmt_rq->params);
 					}
 				}
@@ -1101,10 +1002,9 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 						if (opt->opt_type == PLPGSQL_RAISEOPTION_ERRCODE)
 						{
-							bool		isnull;
 							char	   *value;
 
-							value = plpgsql_check_expr_get_string(cstate, opt->expr, &isnull);
+							value = plpgsql_check_expr_get_string(cstate, opt->expr, NULL);
 
 							if (value != NULL)
 								err_code = plpgsql_check__recognize_err_condition_p(value, true);
@@ -1176,21 +1076,9 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 					if (stmt_execsql->into)
 					{
-
-#if PG_VERSION_NUM >= 110000
-
 						check_variable(cstate, stmt_execsql->target);
 						plpgsql_check_assignment_to_variable(cstate, stmt_execsql->sqlstmt,
 													  stmt_execsql->target, -1);
-
-#else
-
-						plpgsql_check_row_or_rec(cstate, stmt_execsql->row, stmt_execsql->rec);
-						plpgsql_check_assignment(cstate, stmt_execsql->sqlstmt,
-								   stmt_execsql->rec, stmt_execsql->row, -1);
-
-#endif
-
 					}
 					else
 						/* only statement */
@@ -1206,18 +1094,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 									  stmt,
 									  stmt_dynexecute->query,
 									  stmt_dynexecute->into,
-
-#if PG_VERSION_NUM >= 110000
-
 									  stmt_dynexecute->target,
-
-#else
-
-									  stmt_dynexecute->row,
-									  stmt_dynexecute->rec,
-
-#endif
-
 									  stmt_dynexecute->params);
 				}
 				break;
@@ -1226,7 +1103,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 				{
 					PLpgSQL_stmt_open *stmt_open = (PLpgSQL_stmt_open *) stmt;
 					PLpgSQL_var *var = (PLpgSQL_var *) (cstate->estate->datums[stmt_open->curvar]);
-					ListCell	*l;
 
 					plpgsql_check_expr_as_sqlstmt_data(cstate, var->cursor_explicit_expr);
 					plpgsql_check_expr_as_sqlstmt_data(cstate, stmt_open->query);
@@ -1236,11 +1112,14 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 					plpgsql_check_expr_as_sqlstmt_data(cstate, stmt_open->argquery);
 
-					plpgsql_check_expr(cstate, stmt_open->dynquery);
-
-					foreach(l, stmt_open->params)
+					if (stmt_open->dynquery)
 					{
-						plpgsql_check_expr(cstate, (PLpgSQL_expr *) lfirst(l));
+						check_dynamic_sql(cstate,
+										  stmt,
+										  stmt_open->dynquery,
+										  false,
+										  NULL,
+										  stmt_open->params);
 					}
 
 					plpgsql_check_target(cstate, stmt_open->curvar, NULL, NULL);
@@ -1275,7 +1154,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					}
 
 					if (stmt_getdiag->is_stacked &&
-						!is_inside_exception_handler(outer_stmt))
+						!is_inside_exception_handler(outer_stmt_stack))
 					{
 						ereport(ERROR,
 								(errcode(ERRCODE_STACKED_DIAGNOSTICS_ACCESSED_WITHOUT_ACTIVE_HANDLER),
@@ -1289,24 +1168,12 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					PLpgSQL_stmt_fetch *stmt_fetch = (PLpgSQL_stmt_fetch *) stmt;
 					PLpgSQL_var *var = (PLpgSQL_var *) (cstate->estate->datums[stmt_fetch->curvar]);
 
-#if PG_VERSION_NUM >= 110000
-
 					check_variable(cstate, stmt_fetch->target);
 
 					if (var != NULL && var->cursor_explicit_expr != NULL)
 						plpgsql_check_assignment_to_variable(cstate, var->cursor_explicit_expr,
 									   stmt_fetch->target, -1);
 
-#else
-
-					plpgsql_check_row_or_rec(cstate, stmt_fetch->row, stmt_fetch->rec);
-
-					if (var != NULL && var->cursor_explicit_expr != NULL)
-						plpgsql_check_assignment(cstate, var->cursor_explicit_expr,
-									   stmt_fetch->rec, stmt_fetch->row, -1);
-
-#endif
-
 					plpgsql_check_expr(cstate, stmt_fetch->expr);
 
 					cstate->used_variables = bms_add_member(cstate->used_variables, stmt_fetch->curvar);
@@ -1319,8 +1186,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 				break;
 
-#if PG_VERSION_NUM >= 110000
-
 #if PG_VERSION_NUM < 140000
 			case PLPGSQL_STMT_SET:
 				/*
@@ -1338,7 +1203,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 							(errcode(ERRCODE_INVALID_TRANSACTION_TERMINATION),
 							 errmsg("invalid transaction termination")));
 
-				if (is_inside_protected_block(outer_stmt))
+				if (is_inside_protected_block(outer_stmt_stack))
 				{
 					if (stmt->cmd_type == PLPGSQL_STMT_COMMIT)
 						ereport(ERROR,
@@ -1378,8 +1243,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 				}
 				break;
 
-#endif			/* PG_VERSION_NUM >= 110000 */
-
 			default:
 				elog(ERROR, "unrecognized cmd_type: %d", stmt->cmd_type);
 		}
@@ -1389,8 +1252,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
-
-		SPI_restore_connection();
 	}
 	PG_CATCH();
 	{
@@ -1420,9 +1281,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 		}
 
 		MemoryContextSwitchTo(oldCxt);
-
-		/* reconnect spi */
-		SPI_restore_connection();
 	}
 	PG_END_TRY();
 
@@ -1432,6 +1290,37 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 		cstate->was_pragma = false;
 }
 
+static void
+invalidate_strconstvars(PLpgSQL_checkstate *cstate)
+{
+	/*
+	 * We cannot to safely use string constant when we leave related
+	 * path (maybe we can, but it needs deeper analyze ensure so we
+	 * will provess all possible variants).
+	 */
+	if (cstate->top_stmts->invalidate_strconstvars)
+	{
+		int			dno = -1;
+
+		Assert(cstate->strconstvars);
+
+		while ((dno = bms_next_member(cstate->top_stmts->invalidate_strconstvars, dno)) >= 0)
+		{
+			/*
+			 * there is an possibility so string was invalided in some
+			 * nested node. If still it is valid, invalidate it.
+			 */
+			if (cstate->strconstvars[dno])
+			{
+				pfree(cstate->strconstvars[dno]);
+				cstate->strconstvars[dno] = NULL;
+			}
+		}
+
+		pfree(cstate->top_stmts->invalidate_strconstvars);
+	}
+}
+
 /*
  * Ensure check for all statements in list
  *
@@ -1439,17 +1328,23 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 static void
 check_stmts(PLpgSQL_checkstate *cstate, List *stmts, int *closing, List **exceptions)
 {
-	ListCell   *lc;
 	int			closing_local;
 	List	   *exceptions_local;
-	volatile bool		dead_code_alert = false;
 	plpgsql_check_pragma_vector		prev_pragma_vector = cstate->pragma_vector;
+	PLpgSQL_statements current_stmts;
 
 	*closing = PLPGSQL_CHECK_UNCLOSED;
 	*exceptions = NIL;
 
+	current_stmts.outer = cstate->top_stmts;
+	current_stmts.invalidate_strconstvars = NULL;
+	cstate->top_stmts = &current_stmts;
+
 	PG_TRY();
 	{
+		ListCell   *lc;
+		bool		dead_code_alert = false;
+
 		foreach(lc, stmts)
 		{
 			PLpgSQL_stmt	   *stmt = (PLpgSQL_stmt *) lfirst(lc);
@@ -1499,12 +1394,18 @@ check_stmts(PLpgSQL_checkstate *cstate, List *stmts, int *closing, List **except
 				}
 			}
 		}
+
+		invalidate_strconstvars(cstate);
+		cstate->top_stmts = current_stmts.outer;
 	}
 	PG_CATCH();
 	{
 		cstate->pragma_vector = prev_pragma_vector;
 		cstate->was_pragma = false;
 
+		invalidate_strconstvars(cstate);
+		cstate->top_stmts = current_stmts.outer;
+
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -1523,7 +1424,7 @@ push_stmt_to_stmt_stack(PLpgSQL_checkstate *cstate)
 	stmt_stack_item = (PLpgSQL_stmt_stack_item *) palloc0(sizeof(PLpgSQL_stmt_stack_item));
 	stmt_stack_item->stmt = stmt;
 
-	switch (PLPGSQL_STMT_TYPES stmt->cmd_type)
+	switch (stmt->cmd_type)
 	{
 		case PLPGSQL_STMT_BLOCK:
 			stmt_stack_item->label = ((PLpgSQL_stmt_block *) stmt)->label;
@@ -1578,6 +1479,7 @@ pop_stmt_from_stmt_stack(PLpgSQL_checkstate *cstate)
 
 	Assert(cstate->top_stmt_stack != NULL);
 
+
 	cstate->top_stmt_stack = current->outer;
 	pfree(current);
 }
@@ -1588,7 +1490,7 @@ pop_stmt_from_stmt_stack(PLpgSQL_checkstate *cstate)
 static bool
 is_any_loop_stmt(PLpgSQL_stmt *stmt)
 {
-	switch (PLPGSQL_STMT_TYPES stmt->cmd_type)
+	switch (stmt->cmd_type)
 	{
 		case PLPGSQL_STMT_LOOP:
 		case PLPGSQL_STMT_WHILE:
@@ -1636,8 +1538,6 @@ find_nearest_loop(PLpgSQL_stmt_stack_item *current)
 	return NULL;
 }
 
-#if PG_VERSION_NUM >= 110000
-
 /*
  * Returns true, when some outer block handles exceptions.
  * It is used for check of correct usage of COMMIT or ROLLBACK.
@@ -1661,8 +1561,6 @@ is_inside_protected_block(PLpgSQL_stmt_stack_item *current)
 	return false;
 }
 
-#endif
-
 /*
  * This is used for check of correct usage GET STACKED DIAGNOSTICS
  */
@@ -1697,14 +1595,14 @@ found_shadowed_variable(char *varname, PLpgSQL_stmt_stack_item *current, PLpgSQL
 		{
 			PLpgSQL_stmt_block *stmt_block = (PLpgSQL_stmt_block *) current->stmt;
 			int			i;
-			PLpgSQL_datum *d;
 
 			for (i = 0; i < stmt_block->n_initvars; i++)
 			{
 				char	   *refname;
+				PLpgSQL_datum *d;
 
 				d = cstate->estate->func->datums[stmt_block->initvarnos[i]];
-				refname = plpgsql_check_datum_get_refname(d);
+				refname = plpgsql_check_datum_get_refname(cstate, d);
 
 				if (refname != NULL && strcmp(refname, varname) == 0)
 					return true;
@@ -1907,7 +1805,11 @@ dynsql_parser_setup(struct ParseState *pstate, DynSQLParams *params)
 static bool
 has_assigned_tupdesc(PLpgSQL_checkstate *cstate, PLpgSQL_rec *rec)
 {
-	PLpgSQL_rec *target = (PLpgSQL_rec *) (cstate->estate->datums[rec->dno]);
+	PLpgSQL_rec *target;
+
+	Assert(rec);
+
+	target = (PLpgSQL_rec *) (cstate->estate->datums[rec->dno]);
 
 	Assert(rec->dtype == PLPGSQL_DTYPE_REC);
 
@@ -1922,18 +1824,7 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 				  PLpgSQL_stmt *stmt,
 				  PLpgSQL_expr *query,
 				  bool into,
-
-#if PG_VERSION_NUM >= 110000
-
 				  PLpgSQL_variable *target,
-
-#else
-
-				  PLpgSQL_row *row,
-				  PLpgSQL_rec *rec,
-
-#endif
-
 				  List *params)
 {
 	Node	   *expr_node;
@@ -1979,68 +1870,27 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 		if (fexpr->funcid == FORMAT_0PARAM_OID ||
 			fexpr->funcid == FORMAT_NPARAM_OID)
 		{
-			if (fexpr->args && IsA(linitial(fexpr->args), Const))
+			char	   *fmt = NULL;
+			bool		found_ident_placeholder = false;
+			bool		found_literal_placeholder = false;
+			bool		_expr_is_const;
+
+			if (fexpr->args)
+				fmt = plpgsql_check_get_const_string(cstate, linitial(fexpr->args), NULL);
+
+			if (fmt)
 			{
-				StringInfoData	sinfo;
-				char		c, *fmt;
-				bool		subst_is_ok = true;
-				bool		found_ident_placeholder = false;
-				bool		found_literal_placeholder = false;
+				char	   *fstr;
 
-				expr_is_const = fexpr->funcid == FORMAT_0PARAM_OID;
-				fmt = plpgsql_check_const_to_string((Const *) linitial(fexpr->args));
+				fstr = plpgsql_check_get_formatted_string(cstate, fmt, fexpr->args,
+														  &found_ident_placeholder,
+														  &found_literal_placeholder,
+														  &_expr_is_const);
 
-				/*
-				 * The placeholders can be used only in FORMAT_NPARAM function,
-				 * but for simplicity and consistency we check FORMAT_0PARAM and
-				 * FORMAT_NPARAM together
-				 */
-				initStringInfo(&sinfo);
+				/* fix passing 'volatile bool *' to parameter of type 'bool *' discards qualifiers  */
+				expr_is_const = _expr_is_const;
 
-				while ((c = *fmt++))
-				{
-					if (c == '%')
-					{
-						c = *fmt++;
-
-						if (c == '%')
-						{
-							appendStringInfoChar(&sinfo, c);
-						}
-						else if (c == 'I')
-						{
-							appendStringInfoString(&sinfo, "\"%I\"");
-							expr_is_const = false;
-							found_ident_placeholder = true;
-						}
-						else if (c == 'L')
-						{
-							/*
-							 * Original idea was used external parameter,
-							 * but external parameters requires known type,
-							 * so most safe value is NULL instead.
-							 */
-							appendStringInfo(&sinfo, " null ");
-							found_literal_placeholder = true;
-							expr_is_const = false;
-						}
-						else
-						{
-							/*
-							 * Because %s is used, we know nothing about form
-							 * of output string, and has not any sense to continue
-							 * in check.
-							 */
-							subst_is_ok = false;
-							expr_is_const = false;
-							break;
-						}
-					}
-					else
-						appendStringInfoChar(&sinfo, c);
-				}
-
-				if (subst_is_ok)
+				if (fstr)
 				{
 					if (!found_literal_placeholder)
 					{
@@ -2048,26 +1898,26 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 #if PG_VERSION_NUM >= 140000
 
 						/* in this case we can do only basic parser check */
-						raw_parser(sinfo.data, RAW_PARSE_DEFAULT);
+						raw_parser(fstr, RAW_PARSE_DEFAULT);
 
 #else
 
-						raw_parser(sinfo.data);
+						raw_parser(fstr);
 
 #endif
 
 					}
 
 					if (!found_ident_placeholder)
-						dynquery = sinfo.data;
+						dynquery = fstr;
 				}
 			}
 		}
 	}
-	else if (IsA(expr_node, Const))
+	else
 	{
-		expr_is_const = true;
-		dynquery = plpgsql_check_const_to_string((Const *) expr_node);
+		dynquery = plpgsql_check_get_const_string(cstate, expr_node, NULL);
+		expr_is_const = (dynquery != NULL);
 	}
 
 	if (dynquery)
@@ -2079,13 +1929,6 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 
 		dynexpr = palloc0(sizeof(PLpgSQL_expr));
 
-#if PG_VERSION_NUM < 110000
-
-		dynexpr->dtype = PLPGSQL_DTYPE_EXPR;
-		dynexpr->dno = -1;
-
-#endif
-
 #if PG_VERSION_NUM >= 140000
 
 		dynexpr->expr_rw_param = NULL;
@@ -2164,8 +2007,6 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 				RollbackAndReleaseCurrentSubTransaction();
 				MemoryContextSwitchTo(oldCxt);
 				CurrentResourceOwner = oldowner;
-
-				SPI_restore_connection();
 			}
 			PG_CATCH();
 			{
@@ -2219,19 +2060,8 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 			}
 			else if (into)
 			{
-
-#if PG_VERSION_NUM >= 110000
-
 				check_variable(cstate, target);
 				plpgsql_check_assignment_to_variable(cstate, dynexpr, target, -1);
-
-#else
-
-				plpgsql_check_row_or_rec(cstate, row, rec);
-				plpgsql_check_assignment(cstate, dynexpr, rec, row, -1);
-
-#endif
-
 			}
 		}
 
@@ -2284,54 +2114,23 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 		 */
 		if (into && !known_type_of_dynexpr)
 		{
-
-#if PG_VERSION_NUM >= 110000
-
 			if (target->dtype == PLPGSQL_DTYPE_REC)
 				raise_unknown_rec_warning = true;
-
-#else
-
-			if (rec)
-				raise_unknown_rec_warning = true;
-
-#endif
-
 		}
 	}
 
 	/* recheck if target rec var has assigned tupdesc */
 	if (into)
 	{
-
-#if PG_VERSION_NUM >= 110000
+		Assert(target);
 
 		check_variable(cstate, target);
 
 		if (raise_unknown_rec_warning ||
 			(target->dtype == PLPGSQL_DTYPE_REC &&
 			 !has_assigned_tupdesc(cstate, (PLpgSQL_rec *) target)))
-
-#else
-
-		plpgsql_check_row_or_rec(cstate, row, rec);
-
-		if (raise_unknown_rec_warning || (rec != NULL && !has_assigned_tupdesc(cstate, rec)))
-
-#endif
-
 		{
-
-#if PG_VERSION_NUM >= 110000
-
 			if (!bms_is_member(target->dno, cstate->typed_variables))
-
-#else
-
-			if (!bms_is_member(rec->dno, cstate->typed_variables))
-
-#endif
-
 				plpgsql_check_put_error(cstate,
 										0, 0,
 										"cannot determinate a result of dynamic SQL",
